# import numpy as np
# import matplotlib.pyplot as plt
# import math
# from matplotlib.animation import FuncAnimation, PillowWriter

# class Settings:
#     def __init__(self, R1, R2, h, num_sets=1, fig=None, ax=None, speed = 0.5, displacement=10, return_fraction=0.5, grid=False):
#         self.R1 = R1
#         self.R2 = R2
#         self.h = h
#         self.num_sets = num_sets
#         if not fig:
#             fig, ax = plt.subplots()

#             ax.set_xlim(-150 * 3, 150 * 3)
#             ax.set_ylim(-150 - 2 * R1 - 5, 150 + 2 * R1 + 5)
#             ax.set_aspect('equal')
#             # ax.axis('off')
#         self.fig = fig 
#         self.ax = ax
#         if grid:
#             plt.grid(True)

#         interval = 50  # Interval between frames in milliseconds
#         self.speed = speed  # Frequency of the oscillation
#         self.num_teeth = 20  # Number of teeth on each gear
#         self.displacement = displacement
#         self.return_fraction = return_fraction



# class forward:
#     def __init__(self):
#         pass

# class reverse:
#     def __init__(self):
#         pass

# import matplotlib.pyplot as plt
# import numpy as np

# class displacement:
#     def __init__(self):
#         pass

#     def plot(self):
#         # Constants
#         self.Sp = 1  # Unit constant

#         # Generate values for R1/R2 from 0.5 to 2
#         self.R1_R2 = np.linspace(0.5, 5, 500)

#         # Calculate S_R (forward) using the formula S_R = Sp * (R1/R2 - 1)
#         self.S_R_forward = self.Sp * (self.R1_R2 - 1)

#         # Calculate S_R (return) using the formula S_R = Sp * (1 + R1/R2)
#         self.S_R_return = self.Sp * (1 + self.R1_R2)

#         # Calculate L_R using the formula L_R = S_R (forward) + Sp
#         self.L_R = self.S_R_forward + self.Sp

#         # Create a plot
#         self.figure, self.ax = plt.subplots(figsize=(8, 6))
        
#         # Plot S_R (forward)
#         self.ax.plot(self.R1_R2, self.S_R_forward, '--', label=r'$S_R (forward) = S_p \left(\frac{R_1}{R_2} - 1\right)$', color='red')
        
#         # Plot L_R
#         self.ax.plot(self.R1_R2, self.L_R, label=r'$L_R = S_R (forward) + S_p$')
        
#         # Plot S_R (return)
#         self.ax.plot(self.R1_R2, self.S_R_return, label=r'$S_R (return) = S_p \left(1 + \frac{R_1}{R_2}\right)$')
        
#         # Set labels and title
#         self.ax.set_xlabel(r'$\frac{R_1}{R_2}$')
#         self.ax.set_ylabel('Values')
#         self.ax.set_title(r'Plot of $S_R$ (forward), $S_R$ (return), and $L_R$ against $\frac{R_1}{R_2}$')
#         self.ax.grid(True)

#         # Add legend
#         self.ax.legend()
#         return self

#     def S_R_forward(self, S_p, R1, R2):
#         return S_p * (R1 / R2 - 1)

#     def L_forward(self, R1, R2):
#         return self.S_R_forward(self.Sp, R1, R2) + self.Sp


#     def show(self):
#         # Show the plot
#         plt.show()
#         return self

#     def save(self, imagePath):
#         # Save the plot to the specified image path
#         self.figure.savefig(imagePath)
#         return self
    

# class mechanism1:
#     def __init__(self, settings = None):
#         if settings:
#             self.settings = settings
#             self.ax = self.settings.ax
#             self.position = 0
        

#     def info(self):
#         return """
# This mechanism describes a theoretical reciprocating rectilinear gear system designed to achieve a quick return. It converts constant velocity continuous linear motion into reciprocating linear motion of constant velocity with (undesired) instantaneous change of direction.

# Since the input linear velocity is constant, an R1/R2 for the return gear set which gives an SR value which is less than that for the forward stroke is a quick return.
#         """
    
#     def plot(self):
#         self.draw_fixed_racks()
#         self.draw_mobile_rack()
#         return self

#     def draw_fixed_racks(self):
#         y1 = self.settings.h / 2
#         y2 = -self.settings.h / 2
#         x_min = -250
#         x_max = 250

#         # Draw the fixed racks
#         upper_rack_position = y1 + self.settings.R1 + 3 * self.settings.R2
#         lower_rack_position = y2 - self.settings.R1 - 3 * self.settings.R2
#         fixed_rack_upper = self.ax.plot([x_min, x_max], [upper_rack_position, upper_rack_position], 'k-')[0]
#         fixed_rack_lower = self.ax.plot([x_min, x_max], [lower_rack_position, lower_rack_position], 'k-')[0]

#     def draw_mobile_rack_section(self, set_number, stroke="forward/return"):
#         position = self.position
#         y1 = self.settings.h/2
#         ax = self.ax 
#         angle = self.settings.displacement/ self.settings.R2
#         len = angle * self.settings.R1
#         x_min = - len
#         x_max = 0

#         return_section_length = self.settings.displacement * self.settings.return_fraction
#         x_min = x_min if stroke == "forward" else x_max
#         x_max += 0 if stroke == "forward" else return_section_length
#         color = "k" if stroke == "forward" else 'b'
#         mobile_rack_top, = ax.plot([x_min, x_max], [y1, y1], f'{color}-')
#         mobile_rack_bottom, = ax.plot([x_min, x_max], [-y1, -y1], f'{color}-')
#         dotted_line_left, = ax.plot([x_min, x_min], [y1, -y1], f'{color}--')
#         dotted_line_right, = ax.plot([x_max, x_max], [y1, -y1],  f'{color}--')

#     def draw_mobile_rack(self, position=0):
#         num_sets = self.settings.num_sets
#         for i in range(0,num_sets):
#             print(i, num_sets)
#             self.draw_mobile_rack_section(i, stroke = "forward")
#             self.draw_mobile_rack_section(i, "return")
#         pass

#     def show(self):
#         plt.show()
#         return self


#     def save(self, fileName):
#         if self.ani is not None:
#             writer = PillowWriter(fps=25)
#             self.ani.save(f"{fileName}.gif", writer=writer)
#         else:
#             plt.savefig(f"{fileName}.png")
#         return self


import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation

from ....utils import bmutils

class mechanism28(bmutils):
    def __init__(self, settings):
        self.settings = settings
        self.fig, self.ax = plt.subplots(figsize=(10, 6))
        self.ani = None

        # Extract parameters from settings
        self.R1 = self.settings['R1']
        self.R2 = self.settings['R2']
        self.R3 = self.settings['R3']
        self.R4 = self.settings['R4']
        self.R_S = self.settings['R_S']
        self.R_A = self.settings['R_A']
        self.R_B = self.settings['R_B']
        self.R_External = self.settings['R_External']
        self.theta_D = self.settings['theta_D']
        # self.theta_forward = self.settings['theta_forward']
        self.previous_theta = 0
        self.instances = settings["instances"]
        self.phase_6 = settings["phase_6"]
        self.theta_rack = 0
        self.previous_section_type = 0 # it starts with section 0 
        self.current_section_theta = 0
        self.previous_section_theta = 0
        # self.theta_limits = settings["theta_limits"]
        # self.perimeter_theta = settings["perimeter_theta"]
        self.setup_plot()

    def setup_plot(self):
        self.ax.set_title(f'Mechanism 28')
        self.ax.set_xlim(-self.R_B*2, self.R_B*2)
        self.ax.set_ylim(-self.R_B*2, self.R_B*2)
        self.ax.axhline(0, color='black', linewidth=0.5, linestyle='--')
        # self.ax.grid()
        self.ax.axis("off")
        self.ax.set_aspect('equal')
        super().watermark(self.ax)
    
    def plot(self, theta = 0):    
        self.setup_plot()
        # the racks (fixed internal gears)
        x_A, y_A = self.create_circle_section(self.R_A)
        self.ax.plot(x_A, y_A, color='black', linestyle='-.')
        self.ax.plot([0], [0], 'ro') 

        if theta == 0:
            self.previous_section_type = 0
            self.previous_section_theta = 0
            self.previous_theta = 0
            self.current_section_theta = 0 # theta2 = theta3 = theta4
            self.theta_rack = 0
        d_theta_1 = theta - self.previous_theta
        self.previous_theta = theta


        x1, y1 = self.create_circle_section(self.R1, phase=theta)
        self.ax.plot(x1, y1, color='red', linestyle='dotted')
        x1, y1 = self.create_circle_section(self.R1, phase=theta, end_angle=0.2*np.pi) # marker (solid section)
        self.ax.plot(x1, y1, color='red', linestyle='-', linewidth=5)

        
        # check descriptions of these in settings method
        R1 = self.R1
        R2 = self.R2
        R3 = self.R3
        R4 = self.R4
        R_S = self.R_S
        R_A = self.R_A
        R_B = self.R_B
        R_External = self.R_External
       

        ## rotate wheel about
        x_B, y_B = self.create_circle_section(self.R_B, phase=self.phase_6+self.theta_rack, end_angle=self.theta_D)

        segment_instances = self.settings["segment_instances"]
        

        
        '''
        Since there will be just a single instance of these (single planet), adjust the calculation of rack position to use theta_A rather than d_theta_A since sometimes d_theta is greater than max for that section.
        That is the easier way of doing it for a single planet system.
        '''
        for index, instance in enumerate(self.instances):
            self.__dict__.update(instance)
            if theta == 0:
                self.theta = 0
            self.theta += d_theta_1

            ## initialize for each instance
            R2_center =   [R1+R2  , 0]
            R_intersect = [R1+2*R2, 0] # this point is on circumference of R_B

            centerx, centery = self.rotate_points(self.phase_6, [R2_center[0]],[R2_center[1]], [0,0])
            R2_center = [centerx[0],centery[0]]
            centerx, centery = self.rotate_points(self.phase_6, [R_intersect[0]],[R_intersect[1]], [0,0])
            R_intersect = [centerx[0],centery[0]]

            theta_2 = - R1/R2 * self.theta # self.theta is theta of R1
            theta_A = -theta_2 * R3 / R_A #+ self.theta_D/2 # total theta_A (Annulus)
            centerx, centery = self.rotate_points(theta_A, [R2_center[0]],[R2_center[1]], [0,0]) # it is starting at the wrong
            R2_center = [centerx[0],centery[0]]
            self.ax.plot([R2_center[0]], [R2_center[1]], 'ro')

            centerx, centery = self.rotate_points(theta_A, [R_intersect[0]],[R_intersect[1]], [0,0]) # on fixed annulus. 
            R_intersect = [centerx[0],centery[0]]
            # self.ax.plot([R_intersect[0]], [R_intersect[1]], 'ro')
           
            x2, y2 = self.create_circle_section(self.R2, phase=theta_2)
            x3, y3 = self.create_circle_section(self.R3, phase=theta_2)
            self.ax.plot(x2+R2_center[0], y2+R2_center[1], color='red', linestyle='dotted')
            self.ax.plot(x3+R2_center[0], y3+R2_center[1], color='black', linestyle='dotted')

            sections_max_thetas =[
                [segment_instance["section_max_theta"] for segment_instance in segment_instances if segment_instance["type"] == self.previous_section_type][0],
                [segment_instance["section_max_theta"] for segment_instance in segment_instances if segment_instance["type"] != self.previous_section_type][0],
            ]
            section_types = [
                [segment_instance["type"] for segment_instance in segment_instances if segment_instance["type"] == self.previous_section_type][0],
                [segment_instance["type"] for segment_instance in segment_instances if segment_instance["type"] != self.previous_section_type][0],
            ]
            remaining_for_previous_section =  sections_max_thetas[0] - self.previous_section_theta # the angle that we needed to add to fill previous section
            index = 0
            d_theta_2 = R1/R2 * d_theta_1
            previous_section_theta = self.previous_section_theta
            while d_theta_2 > 0 : # we are currently only working with positive d_thetas
                binary_index = (index)%2
                section_theta_max = sections_max_thetas[binary_index]
                remaining_for_previous_section =  section_theta_max - previous_section_theta
                used_in_section = d_theta_2 if d_theta_2 <= remaining_for_previous_section else remaining_for_previous_section
                d_theta_2 -= used_in_section
                total_section_theta = previous_section_theta + used_in_section
                previous_section_theta = total_section_theta if total_section_theta < section_theta_max else 0
                self.previous_section_theta = previous_section_theta

                binary_index_next = (index)%2
                self.previous_section_type =  section_types[binary_index] if previous_section_theta != 0 else section_types[binary_index_next]

                current_section_type = section_types[binary_index]

                # active_segment_type = [segment_instance["type"] for segment_instance in segment_instances if segment_instance["type"] == current_section_type][0]

                if current_section_type == 0: 
                    active_rack_r = R_B
                    active_planet_r = R2
                    not_active_rack_r = R_S
                    not_active_planet_r = R4
                else:
                    active_rack_r = R_S
                    active_planet_r = R4
                    not_active_rack_r = R_B
                    not_active_planet_r = R2
                    # self.ax.plot([R_intersect[0]], [R_intersect[1]], 'ro') 
                    
                d_theta_rack = used_in_section * R3/R_A - used_in_section * active_planet_r/active_rack_r #
                self.theta_rack += d_theta_rack                
                index+=1


            # if instance["type"] == 0:
            self.ax.plot(x2+R2_center[0], y2+R2_center[1], color='green', linestyle=':')
            x2, y2 = self.create_circle_section(self.R2, phase=theta_2, end_angle=0.1*np.pi)
            self.ax.plot(x2+R2_center[0], y2+R2_center[1], color='green', linestyle='-', linewidth=4)
            # else:           
            x4, y4 = self.create_circle_section(self.R4, phase=theta_2)
            self.ax.plot(x4+R2_center[0], y4+R2_center[1], color='blue', linestyle='--')
            x2, y2 = self.create_circle_section(self.R2, phase=theta_2, end_angle=0.1*np.pi)
            self.ax.plot(x2+R2_center[0], y2+R2_center[1], color='red', linestyle='-', linewidth=4)
            
            for key, value in self.__dict__.items():
                if key in instance:  # Check if the key exists in the instance
                    instance[key] = value  # Update the dictionary
                elif hasattr(instance, '__dict__') and key in vars(instance):
                    setattr(instance, key, value)

        segment_endpoints = []
        for instance in segment_instances:
            if instance["type"] == 0:
                active_r = self.R_B
                color = "green"
            else:
                active_r = self.R_S
                color = "blue"
            x_i, y_i = self.create_circle_section(active_r, phase=self.phase_6+self.theta_rack + instance["theta"], end_angle=self.theta_D)
            instance["segment"] = [x_i, y_i]
            self.ax.plot(x_i, y_i, color=color, linestyle='--',)
            
            segment_endpoints.append([x_i[-1], y_i[-1]])
            segment_endpoints.append([x_i[0], y_i[0]])

        for index, point in enumerate(segment_endpoints): # the lines conneting end of R_B to beginning of R_S
            if index%2 == 0:
                continue
            next_point_index = index+1 if (index+1) < len(segment_endpoints) else 0
            next_point = segment_endpoints[next_point_index]

            # Draw lines between the points
            self.ax.plot([point[0],next_point[0]],[point[1],next_point[1]] , color='black') 
            # break
            
        x_External, y_External = self.create_circle_section(R_External, phase=self.phase_6+self.theta_rack)
        self.ax.plot(x_External, y_External, color='black',linestyle='--')
        
        # ''' Markers '''
        x_External, y_External = self.create_circle_section(R_External, phase=self.phase_6+self.theta_rack, end_angle=0.1*np.pi)
        self.ax.plot(x_External, y_External, color='black', linestyle='-', linewidth=5)
        
        return self
    
    def update(self, interval=100, frames=100):
        speed = self.settings["speed"]
        len = self.settings["animation_rounds"]
        displacements = np.linspace(0, len, int(250/speed))
        # displacements = np.concatenate([displacements, displacements[::-1]])  # Reverse motion
        def animate(displacement):
            self.ax.clear()
            self.plot(displacement)

        
        self.setup_plot()
        self.ani = animation.FuncAnimation(self.fig, lambda displacement: animate(displacement), frames=displacements, interval=50, repeat=True)
        return self
        

    @staticmethod
    # return_speed: how many times return is faster than forward
    # frequency: number of oscillations of output per revolution of planet. Should be greater than one
    def Settings(R1=5,R2=20, return_speed=2, frequency = 1, animation_rounds=1, speed=1):
        num_hands = frequency *2
        theta_segment = 2*np.pi/num_hands # angle of rack segment used
        theta_D = theta_segment # theta_disk

        k = return_speed

        R_B = R1 + 2 * R2 # R_Big
        theta_2_R = theta_D * R_B / R2  # Disk 2 rotates through theta_2_R to cover distance of theta_D * R_B(entire length of rack)

        theta_D_R  = (1/(k+1) ) * theta_D
        theta_D_F = 2 * theta_D - theta_D_R

        theta_3_R = theta_2_R
        c = theta_D_R / theta_3_R
        R3 = (c * R1 + c * R2)/(1-c)
        R_A = R1 + R2 + R3  # R_Annulus
        

        theta_3_F = theta_D_F * R_A / R3
        theta_4_F = theta_3_F
        c = theta_D / theta_4_F

        R4 = c*(R1 + R2) / (1-c)
        R_S = R1 + R2 + R4 # R_Small. Moves mechanism forward in contrast to R_Big which moves it backwards

        R_External = R_B * 1.2

        theta_3_perimeter = 2*np.pi * R_A / R3
        s_2_perimeter = theta_3_perimeter * R2
        theta_1_perimeter = s_2_perimeter / R1
        perimeter_theta = theta_1_perimeter
        animation_rounds = int(animation_rounds * perimeter_theta)
        num_instances = num_hands
        theta_points = [(perimeter_theta/num_instances) * i for i in range(num_instances)]
        instances = [{"theta":0}]
        # for sections on R_B(ig) and R_S(mall)
        s_section_B = theta_D * R_B
        s_section_S = theta_D * R_S
        theta_max_B = s_section_B/R2 # both theta2 and theta3 since they are equal, being co-axial
        theta_max_S = s_section_S/R4 # both theta4 and theta2
        section_max_theta = [theta_max_B,theta_max_S ] # theta_3s
        segment_instances = [{"theta":theta, "type":index%2, "section_max_theta":section_max_theta[index%2]} for index, theta in enumerate(theta_points)]
        for index, instance in enumerate(instances):
            instance["previous_direction"] = 1

        phase_6 = - theta_D_R/2 - np.pi/2
        return {
            "R1": R1,
            "R2": R2,
            "R3": R3,
            "R4": R4,
            "R_S": R_S,
            "R_A": R_A,
            "R_B": R_B,
            "R_External": R_External,
            "theta_D": theta_D,
            # "theta_forward":theta_forward,
            "animation_rounds":animation_rounds,
            "speed":speed,
            "perimeter_theta":perimeter_theta,
            # "theta_limits":theta_limits,
            "instances":instances,
            "segment_instances":segment_instances,
            "phase_6":phase_6,
            "frequency":frequency,
        }
    @staticmethod
    def info():
        description = (
            "Mechanism Description:\n"
            "-----------------------\n"
            "Continuous rotary to reciprocating rotary and/or linear motion with possible configuration for quick return.\n"
            "**Applications**:\n"
            "This type of mechanism is intended to be used in the development of clamp-on self-actuated long-range linear motion mechanisms.\n"
        )
        return description


import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation

from ....utils import bmutils
import time
class mechanism31(bmutils):
    def __init__(self, settings):
        self.settings = settings
        self.fig, self.ax = plt.subplots(figsize=(10, 6))
        self.ani = None

        # Extract parameters from settings
        self.R1 = self.settings['R1']
        self.R2 = self.settings['R2']
        self.R3 = self.settings['R3']
        self.R4 = self.settings['R4']
        self.R_S = self.settings['R_S']
        self.R_A = self.settings['R_A']
        self.R_B = self.settings['R_B']
        self.R_External = self.settings['R_External']
        self.theta_D = self.settings['theta_D']
        # self.theta_forward = self.settings['theta_forward']
        # self.previous_theta = 0
        self.instances = settings["instances"]
        self.phase_6 = settings["phase_6"]
        
        self.setup_plot()

    def setup_plot(self):
        self.ax.set_title(f'Mechanism 31')
        self.ax.set_xlim(-self.R_B*2, self.R_B*2)
        self.ax.set_ylim(-self.R_B*2, self.R_B*2)
        self.ax.axhline(0, color='black', linewidth=0.5, linestyle='--')
        # self.ax.grid()
        self.ax.axis("off")
        self.ax.set_aspect('equal')
        super().watermark(self.ax)
    
    def plot(self, theta = 0):    
        self.setup_plot()
        # the racks (fixed internal gears)
        x_A, y_A = self.create_circle_section(self.R_A)
        self.ax.plot(x_A, y_A, color='black', linestyle='-.')
        self.ax.plot([0], [0], 'ro') 

        # if theta == 0:
        #     self.previous_theta = 0
        #     self.current_section_theta = 0 # theta2 = theta3 = theta4
        # d_theta_1 = theta - self.previous_theta
        # self.previous_theta = theta


        x1, y1 = self.create_circle_section(self.R1, phase=theta)
        self.ax.plot(x1, y1, color='red', linestyle='dotted')
        x1, y1 = self.create_circle_section(self.R1, phase=theta, end_angle=0.2*np.pi) # marker (solid section)
        self.ax.plot(x1, y1, color='red', linestyle='-', linewidth=5)

        
        # check descriptions of these in settings method
        R1 = self.R1
        R2 = self.R2
        R3 = self.R3
        R4 = self.R4
        R_S = self.R_S
        R_A = self.R_A
        R_B = self.R_B
        R_External = self.R_External
       

        ## rotate wheel about
        # x_B, y_B = self.create_circle_section(self.R_B, phase=self.phase_6+self.theta_rack, end_angle=self.theta_D)

        segment_instances = self.settings["segment_instances"]

        def get_previous_section_type(prev_theta, theta_2_initial = 0):
            sections_max_thetas =[
                [segment_instance["section_max_theta"] for segment_instance in segment_instances if segment_instance["type"] == 0][0], # since we are starting from theta = 0
                [segment_instance["section_max_theta"] for segment_instance in segment_instances if segment_instance["type"] != 0][0],
            ]
            section_types = [
                [segment_instance["type"] for segment_instance in segment_instances if segment_instance["type"] == 0][0],
                [segment_instance["type"] for segment_instance in segment_instances if segment_instance["type"] != 0][0],
            ]
            rack_theta = 0
            d_theta_2 = R1/R2 * prev_theta
            current_section_type = 0
            # previous_section_type = 0
            previous_section_theta = 0
            index_while = 0
            while d_theta_2 > 0 : # we are currently only working with positive d_thetas
                binary_index = (index_while)%2
                section_theta_max = sections_max_thetas[binary_index]
                remaining_for_previous_section =  section_theta_max - previous_section_theta
                used_in_section = d_theta_2 if d_theta_2 <= remaining_for_previous_section else remaining_for_previous_section
                
                total_section_theta = previous_section_theta + used_in_section
                previous_section_theta = total_section_theta if total_section_theta < section_theta_max else 0

                
                binary_index_next = (index_while+1)%2
                # previous_section_type =  section_types[binary_index] if previous_section_theta != 0 else section_types[binary_index_next]
                # if index == 1:
                # print(index, binary_index, prev_theta, d_theta_2, section_theta_max, used_in_section, remaining_for_previous_section, total_section_theta, previous_section_theta, current_section_type)

                current_section_type = section_types[binary_index]

                # active_segment_type = [segment_instance["type"] for segment_instance in segment_instances if segment_instance["type"] == current_section_type][0]

                if current_section_type == 0: 
                    active_rack_r = R_B
                    active_planet_r = R2
                else:
                    active_rack_r = R_S
                    active_planet_r = R4
                    # self.ax.plot([R_intersect[0]], [R_intersect[1]], 'ro') 
                    
                # d_theta_rack = used_in_section * R3/R_A - used_in_section * active_planet_r/active_rack_r #
                # for section with offset, we need to subtract the initial offset
                d_theta_rack = used_in_section * R3/R_A - used_in_section * active_planet_r/active_rack_r #
                # print(d_theta_rack)
                rack_theta += d_theta_rack   
                
                d_theta_2 -= used_in_section
                index_while += 1
            # print(index, previous_section_type,current_section_type, prev_theta)
            # theta_rack_initial = 0
            # if theta_2_initial > 0:
            #     theta_rack_initial = get_previous_section_type(theta_2_initial, 0) # we need to convert theta initial to theta rack so that we get where the rack started
            #     print(theta_rack_initial, theta_2_initial)
            
            rack_theta -= theta_2_initial * R3/R_A 
                # rack_theta += theta_rack_initial #* R3/R_A 
            return rack_theta
    
        def theta_2_to_theta_1(theta_2): # will be same value if R2 = R1
            return theta_2 * R2/R1

        for index, instance in enumerate(self.instances):
            self.__dict__.update(instance)
            # if theta == 0:
            #     self.theta = theta_2_to_theta_1(self.theta_2_init) # theta for this instance
            #     # theta_init is in terms of theta_2 need to conver to theta_1

            self.theta = theta_2_to_theta_1(self.theta_2_init) + theta
            self.theta_rack = get_previous_section_type(self.theta, theta_2_to_theta_1(self.theta_2_init)) # what bug is it here??? use so that we do not need to set theta_rack_init and previous_section_type_init
            # print(index)
            # if index ==1:
            #     time.sleep(1)
            

            ## initialize for each instance
            
            R_intersect = [R1+2*R2, 0] # this point is on circumference of R_B # No longer needed. Was used by previus very bad algorithm

            '''
            Driver parts
            '''
            R2_center =   [R1+R2  , 0]
            
            centerx, centery = self.rotate_points(self.phase_6, [R2_center[0]],[R2_center[1]], [0,0])
            R2_center = [centerx[0],centery[0]]
            centerx, centery = self.rotate_points(self.phase_6, [R_intersect[0]],[R_intersect[1]], [0,0])
            R_intersect = [centerx[0],centery[0]]

            theta_2 = - R1/R2 * self.theta # self.theta is theta of R1
            theta_A = -theta_2 * R3 / R_A #+ self.theta_D/2 # total theta_A (Annulus)
            centerx, centery = self.rotate_points(theta_A, [R2_center[0]],[R2_center[1]], [0,0]) # it is starting at the wrong
            R2_center = [centerx[0],centery[0]]
            if index == 0:
                self.ax.plot([R2_center[0]], [R2_center[1]], 'ro')

            centerx, centery = self.rotate_points(theta_A, [R_intersect[0]],[R_intersect[1]], [0,0]) # on fixed annulus. 
            R_intersect = [centerx[0],centery[0]]
            # self.ax.plot([R_intersect[0]], [R_intersect[1]], 'ro')
           
            if index == 0:
                x2, y2 = self.create_circle_section(self.R2, phase=theta_2)
                x3, y3 = self.create_circle_section(self.R3, phase=theta_2)
                self.ax.plot(x2+R2_center[0], y2+R2_center[1], color='red', linestyle='dotted')
                self.ax.plot(x3+R2_center[0], y3+R2_center[1], color='black', linestyle='dotted')


            # if instance["type"] == 0:
            if index == 0:
                self.ax.plot(x2+R2_center[0], y2+R2_center[1], color='green', linestyle=':')
                x2, y2 = self.create_circle_section(self.R2, phase=theta_2, end_angle=0.1*np.pi)
                self.ax.plot(x2+R2_center[0], y2+R2_center[1], color='green', linestyle='-', linewidth=4)
                # else:           
                x4, y4 = self.create_circle_section(self.R4, phase=theta_2)
                self.ax.plot(x4+R2_center[0], y4+R2_center[1], color='blue', linestyle='--')
                x2, y2 = self.create_circle_section(self.R2, phase=theta_2, end_angle=0.1*np.pi)
                self.ax.plot(x2+R2_center[0], y2+R2_center[1], color='red', linestyle='-', linewidth=4)
            '''
            End of driver parts
            '''

            for key, value in self.__dict__.items():
                if key in instance:  # Check if the key exists in the instance
                    instance[key] = value  # Update the dictionary
                elif hasattr(instance, '__dict__') and key in vars(instance):
                    setattr(instance, key, value)

            
            '''
            Driven parts (output)
            '''

            # ''' Markers '''
            # print(index, self.theta_rack)

            segment_endpoints = []
            for _instance in segment_instances:
                color = None
                if "color" in instance.keys():
                    color = instance["color"]
            
                if _instance["type"] == 0:
                    active_r = self.R_B
                    color = color if color else ("green" if index == 0 else "orange")
                else:
                    active_r = self.R_S
                    color = color if color else ("green" if index == 0 else "orange")

                x_i, y_i = self.create_circle_section(active_r, phase=self.phase_6+self.theta_rack + _instance["theta"], end_angle=self.theta_D)
                instance["segment"] = [x_i, y_i]
                self.ax.plot(x_i, y_i, color=color, linestyle='--',)
                
                segment_endpoints.append([x_i[-1], y_i[-1]])
                segment_endpoints.append([x_i[0], y_i[0]])

            for index, point in enumerate(segment_endpoints): # the lines conneting end of R_B to beginning of R_S
                if index%2 == 0:
                    continue
                next_point_index = index+1 if (index+1) < len(segment_endpoints) else 0
                next_point = segment_endpoints[next_point_index]

                # Draw lines between the points
                self.ax.plot([point[0],next_point[0]],[point[1],next_point[1]] , color=color) 

            angle = np.arctan2(R_intersect[1], R_intersect[0])
            R_CAM = self.settings["R_CAM"]
            num_segments = len(self.settings["segment_instances"])
            num_segments /= 2
            for index in range(int(num_segments)):
                x_CAM, y_CAM = self.create_circle_section(R_CAM, phase=angle-2*np.pi/(num_segments*2)-self.settings["theta_CAM"]/2+(2*np.pi/num_segments)*index, end_angle=self.settings["theta_CAM"])
                self.ax.plot(x_CAM, y_CAM, color=color, linestyle='-')
                start_CAM = (x_CAM[0], y_CAM[0])  # Start of the segment
                end_CAM = (x_CAM[-1], y_CAM[-1])  # End of the segment
                # Draw a line from start to end of the segment
                self.ax.plot([start_CAM[0], end_CAM[0]], [start_CAM[1], end_CAM[1]], color=color)

            x_External, y_External = self.create_circle_section(R_External, phase=self.phase_6+self.theta_rack, end_angle=0.1*np.pi)
            self.ax.plot(x_External, y_External, color=color, linestyle='-', linewidth=5)
            # break
            
        x_External, y_External = self.create_circle_section(R_External, phase=self.phase_6+self.theta_rack)
        self.ax.plot(x_External, y_External, color='black',linestyle='--')
        
        
        return self
    
    def update(self, interval=100, frames=100):
        speed = self.settings["speed"]
        len = self.settings["animation_rounds"]
        displacements = np.linspace(0, len, int(250/speed))
        # displacements = np.concatenate([displacements, displacements[::-1]])  # Reverse motion
        def animate(displacement):
            self.ax.clear()
            self.plot(displacement)

        
        self.setup_plot()
        self.ani = animation.FuncAnimation(self.fig, lambda displacement: animate(displacement), frames=displacements, interval=50, repeat=True)
        return self
        

    @staticmethod
    # return_speed: how many times return is faster than forward
    # frequency: number of oscillations of output per revolution of planet. Should be greater than one
    def Settings(R1=5,R2=20, return_speed=2, frequency = 1, num_instances =2, phase_difference = 0, animation_rounds=1, speed=1):
        phase_difference_fraction = phase_difference/2*np.pi
        num_hands = frequency *2
        theta_segment = 2*np.pi/num_hands # angle of rack segment used
        theta_D = theta_segment # theta_disk

        k = return_speed

        R_B = R1 + 2 * R2 # R_Big
        theta_2_R = theta_D * R_B / R2  # Disk 2 rotates through theta_2_R to cover distance of theta_D * R_B(entire length of rack)

        theta_D_R  = (1/(k+1) ) * theta_D
        theta_D_F = 2 * theta_D - theta_D_R

        theta_3_R = theta_2_R
        c = theta_D_R / theta_3_R
        R3 = (c * R1 + c * R2)/(1-c)
        R_A = R1 + R2 + R3  # R_Annulus
        

        theta_3_F = theta_D_F * R_A / R3
        theta_4_F = theta_3_F
        c = theta_D / theta_4_F

        R4 = c*(R1 + R2) / (1-c)
        R_S = R1 + R2 + R4 # R_Small. Moves mechanism forward in contrast to R_Big which moves it backwards

        R_External = R_B * 1.2

        R_CAM = R_B * 1.4

        theta_3_perimeter = 2*np.pi * R_A / R3
        s_2_perimeter = theta_3_perimeter * R2
        theta_1_perimeter = s_2_perimeter / R1
        perimeter_theta = theta_1_perimeter
        animation_rounds = int(animation_rounds * perimeter_theta)
        theta_points = [(perimeter_theta/num_hands) * i for i in range(num_hands)]
        
        # for sections on R_B(ig) and R_S(mall)
        s_section_B = theta_D * R_B # this is theta 2 since s_section_B x R2 = theta_D * R_B
        s_section_S = theta_D * R_S
        theta_max_B = s_section_B/R2 # both theta2 and theta3 since they are equal, being co-axial
        theta_max_S = s_section_S/R4 # both theta4 and theta2
        section_max_theta = [theta_max_B,theta_max_S ] # theta_3s
        theta_CAM = 0.75 * theta_D_F


        # how to calculate theta_2_init from phase_difference_fraction 
        R_A_difference = phase_difference_fraction
        fraction = 0.5
        # print(theta_max_B*R2/R_A)
        theta_3_init = R_A * phase_difference_fraction/R3
        theta_2_init = theta_3_init
        instances = [
            {"theta":0, "theta_rack":0, "theta_2_init":0, "color":"black"},
            {"theta":0, "theta_rack":0, "theta_2_init":theta_2_init, "color":"orange"},
            # {"theta":0, "theta_rack":0, "theta_2_init":0, "color":"black"},
            # #  {"theta":0, "theta_rack":0, "theta_init":0, "color":"black"},
            
            # {"theta":s_section_S*fraction, "theta_init":s_section_S*fraction, "theta_rack":-theta_D*fraction,"theta_rack_init":-theta_D*fraction}
        ] # the second one starts at end of first B
        colors = [
           "","", "green", "gray"
        ]
        if num_instances > 2:
            for i in range(2,num_instances):
                instance = instances[i%2].copy()
                instance["theta_2_init"] = theta_2_init * (i)
                instance["color"] = colors[i]
                instances.append(instance)

        segment_instances = [{"theta":theta, "type":index%2, "section_max_theta":section_max_theta[index%2]} for index, theta in enumerate(theta_points)]
        # for index, instance in enumerate(instances):
        #     instance["previous_direction"] = 1

        phase_6 = - theta_D_R/2 - np.pi/2
        return {
            "R1": R1,
            "R2": R2,
            "R3": R3,
            "R4": R4,
            "R_S": R_S,
            "R_A": R_A,
            "R_B": R_B,
            "R_External": R_External,
            "R_CAM": R_CAM,
            "theta_CAM": theta_CAM,
            "theta_D": theta_D,
            # "theta_forward":theta_forward,
            "animation_rounds":animation_rounds,
            "speed":speed,
            "perimeter_theta":perimeter_theta,
            # "theta_limits":theta_limits,
            "instances":instances,
            "segment_instances":segment_instances,
            "phase_6":phase_6,
            "frequency":frequency,
        }
    @staticmethod
    def info():
        description = (
            "Mechanism Description:\n"
            "-----------------------\n"
            "Mechanism 30 with cams .\n"
            "**Applications**:\n"
            "This type of mechanism is intended to be used in the development of clamp-on self-actuated long-range linear motion mechanisms and walking mechanisms.\n"
        )
        return description


import matplotlib.pyplot as plt
import numpy as np
import math
from sklearn.linear_model import LinearRegression

class RR:
    def __init__(self, vars=None):
        self.θ_0 = 0 # default value
        self.θ_0_increment = 0
        self.vertical_threshold = 0.01
        self.theta_values = np.linspace(-np.pi/2, 1.5 * np.pi, 400)
        self.fig = self.figH = self.figS = None
        self.axs = None
        self.animes = []
        self.vars = vars
        if vars is not None:
            self.setVars(vars)
        if self.outputType == "plot": # static
          if self.plotType == "locus":
            self.plotAll(self.outputOptions)
          elif self.plotType =="values":
            self.plotallvalues(self.outputOptions)
          plt.show()
        elif self.outputType == "animation": #nimation
          plt.ioff()
          if self.plotType == "locus":
            self.plotAll(self.outputOptions)
          elif self.plotType =="values":
            self.plotallvalues(self.outputOptions)
          self.funcAnimation()


    def setVars(self, vars):
        self.__dict__.update(vars)

    def curve_parametric(self):
        θ_0 = self.θ_0 + self.θ_0_increment
        theta = self.theta_values
        R = self.R
        r = self.R * self.f
        a = self.a
        x = R * np.cos(theta) + r * np.cos((1 + a) * theta + math.radians(θ_0))
        y = R * np.sin(theta) + r * np.sin((1 + a) * theta + math.radians(θ_0))
        return x, y

    def curvature_parametric(self, R, f, a, θ_0_increment=False):
        θ_0 = self.θ_0 + (self.θ_0_increment if θ_0_increment == False else θ_0_increment)
        theta = self.theta_values
        r = R * f
        dx_dtheta = -R * np.sin(theta) - r * (1 + a) * np.sin((1 + a) * theta + math.radians(θ_0))
        dy_dtheta = R * np.cos(theta) + r * (1 + a) * np.cos((1 + a) * theta + math.radians(θ_0))
        d2x_dtheta2 = -R * np.cos(theta) - r * (1 + a)**2 * np.cos((1 + a) * theta + math.radians(θ_0))
        d2y_dtheta2 = -R * np.sin(theta) - r * (1 + a)**2 * np.sin((1 + a) * theta + math.radians(θ_0))
        curvature = (dx_dtheta * d2y_dtheta2 - dy_dtheta * d2x_dtheta2) / (dx_dtheta**2 + dy_dtheta**2)**(3/2)
        return dx_dtheta, dy_dtheta, curvature

    def __plotCurve(self, plot):
        self.setRequiredTheta0()

        x_curve, y_curve = self.curve_parametric()
        longest_segment = self.get_longest_straight_line()
        plot.clear()
        plot.plot(x_curve, y_curve)
        plot.plot(x_curve[longest_segment], y_curve[longest_segment], color='green', label='Longest Segment', linewidth=8)
        plot.set_title(rf'Output Locus(θ$_0$: {self.getRequiredTheta0()}°, h:{self.getdH():.4f}, s:{self.getdS():.4f}, l:{self.getStraightSectionFraction():.2f}%)')
        plot.set_aspect('equal')

    def plotStraightSection(self, plot):
        # the straight line section
        x_straight, y_reg = self.expected_straight_line()
        plot.plot(x_straight, y_reg, color='red', linestyle='--', label='Linear Regression')
    def plotCurvature(self, plot):
        _, _, curvature = self.curvature_parametric(self.R, self.f, self.a)
        theta_degrees = np.degrees(self.theta_values)
        longest_segment = self.get_longest_straight_line()
        # curvature and straight segment
        plot.clear()
        plot.plot(theta_degrees, curvature)
        plot.set_title(f'Plot of Curvature(max:{self.getmaxCurvature():.6f})')

        plot.scatter(theta_degrees[longest_segment], curvature[longest_segment], color='red')
        # plot.legend()
    def plotdS(self, plot):
        pass
    def plotdH(self, plot):
        pass
    def plotmaxCurvature(self, plot):
        pass
    def plotStraightSectionFraction(self, plot):
        pass

    def plotAll(self, options=[]):
        # numGraphs = len(options.keys())
        numGraphs = len(options)
        # numGraphs = numGraphs if numGraphs > 1 else numGraphs + 1
        fig = axs = None

        if self.fig == None:
          fig, axs = plt.subplots(1, numGraphs, figsize=(12, 6))
          self.fig = fig
          self.axs = axs
        else:
          fig = self.fig
          axs = self.axs
        self.fig.suptitle(f"Straight Line Generator (R:{self.R:.2f}, f:{self.f:.2f}, a:{self.a}, curvature:{self.curvature_threshold:.2f})")
        axsIndex = 0
        for key in options:
          if key == "curve":
            try:
              self.__plotCurve(axs[axsIndex])
              self.plotStraightSection(axs[axsIndex])
              axsIndex+=1
            except: # just a single axs
              self.__plotCurve(axs)
              self.plotStraightSection(axs)
          if key == "curvature":
            try:
              self.plotCurvature(axs[axsIndex])
              axsIndex+=1
            except:
              self.plotCurvature(axs)
          if key == "curvatures":
            pass

        plt.tight_layout()
        # plt.show()
    def updateAllPlot(self, frame, numFrames, updateVar, plotOptions):
        for var_name, var_data in updateVar.items():
          if var_name == "numFrames":
            continue
          if var_name == 'R':
              R = var_data['min'] + (var_data['max'] - var_data['min']) * frame / numFrames
              self.R = R
              self.fig.suptitle(f"Straight Line Generator (R) (R:{self.R:.2f}, f:{self.f:.2f}, a:{self.a}, curvature:{self.curvature_threshold:.2f})")
          elif var_name == 'f':
              f = var_data['min'] + (var_data['max'] - var_data['min']) * frame / numFrames
              self.f = f
              self.fig.suptitle(f"Straight Line Generator (f) (R:{self.R:.2f}, f:{self.f:.2f}, a:{self.a}, curvature:{self.curvature_threshold:.2f})")
          elif var_name == 'a':
              a = int(var_data['min'] + (var_data['max'] - var_data['min']) * frame / numFrames)
              self.a = a
              self.fig.suptitle(f"Straight Line Generator (a) (R:{self.R:.2f}, f:{self.f:.2f}, a:{self.a}, curvature:{self.curvature_threshold:.2f})")
          elif var_name == 'curvature_threshold':
              curvature_threshold = var_data['min'] + (var_data['max'] - var_data['min']) * frame / numFrames
              self.curvature_threshold = curvature_threshold
              self.fig.suptitle(f"Straight Line Generator (curvature) (R:{self.R:.2f}, f:{self.f:.2f}, a:{self.a}, curvature:{self.curvature_threshold:.2f})")
          break # do for only the first

        axsIndex = 0
        for key in plotOptions:
          if key == "curve":
            try:
              self.__plotCurve(self.axs[axsIndex])
              self.plotStraightSection(self.axs[axsIndex])
              axsIndex+=1
            except: # just a single axs
              self.__plotCurve(self.axs)
              self.plotStraightSection(self.axs)
          if key == "curvature":
            try:
              self.plotCurvature(self.axs[axsIndex])
              axsIndex+=1
            except:
              self.plotCurvature(self.axs)
          if key == "curvatures":
            pass


    def updateAllPlotForValues(self, frame, numFrames, updateVar, plotOptions):
        for var_name, var_data in updateVar.items():
          if var_name == "numFrames":
            continue
          if var_name == 'R':
              R = var_data['min'] + (var_data['max'] - var_data['min']) * frame / numFrames
              self.R = R
              self.figH.suptitle(f"Straight Line Generator (R) (R:{self.R:.2f}, f:{self.f:.2f}, a:{self.a}, curvature:{self.curvature_threshold:.2f})") if self.figH is not None else None
              self.figS.suptitle(f"Straight Line Generator (R) (R:{self.R:.2f}, f:{self.f:.2f}, a:{self.a}, curvature:{self.curvature_threshold:.2f})") if self.figS is not None else None
          elif var_name == 'f':
              f = var_data['min'] + (var_data['max'] - var_data['min']) * frame / numFrames
              self.f = f
              self.figH.suptitle(f"Straight Line Generator (f) (R:{self.R:.2f}, f:{self.f:.2f}, a:{self.a}, curvature:{self.curvature_threshold:.2f})") if self.figH is not None else None
              self.figS.suptitle(f"Straight Line Generator (f) (R:{self.R:.2f}, f:{self.f:.2f}, a:{self.a}, curvature:{self.curvature_threshold:.2f})") if self.figS is not None else None
          elif var_name == 'a':
              a = int(var_data['min'] + (var_data['max'] - var_data['min']) * frame / numFrames)
              self.a = a
              self.figH.suptitle(f"Straight Line Generator (a) (R:{self.R:.2f}, f:{self.f:.2f}, a:{self.a}, curvature:{self.curvature_threshold:.2f})") if self.figH is not None else None
              self.figS.suptitle(f"Straight Line Generator (a) (R:{self.R:.2f}, f:{self.f:.2f}, a:{self.a}, curvature:{self.curvature_threshold:.2f})") if self.figS is not None else None
          elif var_name == 'curvature_threshold':
              curvature_threshold = var_data['min'] + (var_data['max'] - var_data['min']) * frame / numFrames
              self.curvature_threshold = curvature_threshold
              self.figH.suptitle(f"Straight Line Generator (curvature) (R:{self.R:.2f}, f:{self.f:.2f}, a:{self.a}, curvature:{self.curvature_threshold:.2f})") if self.figH is not None else None
              self.figS.suptitle(f"Straight Line Generator (curvature) (R:{self.R:.2f}, f:{self.f:.2f}, a:{self.a}, curvature:{self.curvature_threshold:.2f})") if self.figS is not None else None
          
          if self.figH is not None:
            self.plotallvalues(["h"])
          if self.figS is not None:
            self.plotallvalues(["s"])
          break # do for only the first

    def expected_straight_line(self):
        longest_segment = self.get_longest_straight_line()
        x_straight = self.curve_parametric()[0][longest_segment]
        y_straight = self.curve_parametric()[1][longest_segment]
        ret_x=ret_y = None
        # Check if the line is nearly vertical because linear regression does not work so well with vertical lines
        if len(x_straight) == 0:
          return (0,0)
        if np.abs(np.max(x_straight) - np.min(x_straight)) < self.vertical_threshold:  # Adjust threshold as needed
            # Handle vertical line by using a fixed x-coordinate
            x_value = np.mean(x_straight)  # You can choose mean, median, or another central value
            x_straight[:] = x_value
            ret_x = x_straight
            ret_y = y_straight
            self.slope = np.Infinity
        # Fit linear regression to the straight section (for non-vertical lines)
        else:
            reg = LinearRegression().fit(x_straight[:, np.newaxis], y_straight)
            y_reg = reg.predict(x_straight[:, np.newaxis])
            ret_x = x_straight
            ret_y = y_reg
            # return (x_straight, y_reg)
            slope = (ret_y[-1] - ret_y[0]) / (ret_x[-1] - ret_x[0])
            self.slope = slope
            angle_radians = np.arctan(slope)
            self.lineAngle = np.degrees(angle_radians)
            # print(self.lineAngle)

        return (ret_x, ret_y)

    def get_longest_straight_line(self):
        _,_,curvature = self.curvature_parametric(self.R, self.f, self.a)
        straight_line_indices = np.where(np.abs(curvature) < self.curvature_threshold)[0]
        segments = np.split(straight_line_indices, np.where(np.diff(straight_line_indices) != 1)[0] + 1)

        # Check if there are segments both at the beginning and end
        if segments and len(segments) > 1 and segments[-1][-1] == len(curvature) - 1 and segments[0][0] == 0:
            # Combine the segments into the first segment
            combined_segment = np.concatenate((segments[-1], segments[0]))
            # Remove the last segment
            segments.pop()
            # Replace the first segment with the combined segment
            segments[0] = combined_segment

        longest_segment = max(segments, key=len)
        return longest_segment

    def setRequiredTheta0(self):
      self.θ_0_increment = self.getRequiredθ_0_increment()
    def getRequiredTheta0(self):
      # θ_0_increment = self.getRequiredθ_0_increment() # because this function will be called after setRequiredTheta0
      return self.θ_0_increment + self.θ_0

    def middle_of_line_at_0(self):
        count=0;
        θ_0_increment = 0
        θ_0_increment_interval = 1
        _,_,curvature = self.curvature_parametric(self.R, self.f, self.a, θ_0_increment)
        while count < len(curvature)//θ_0_increment_interval: # mat
          θ_0_increment += θ_0_increment_interval
          _,_,curvature = self.curvature_parametric(self.R, self.f, self.a, θ_0_increment)
          straight_line_indices = np.where(np.abs(curvature) < self.curvature_threshold)[0]
          segments = np.split(straight_line_indices, np.where(np.diff(straight_line_indices) != 1)[0] + 1)

          if segments and len(segments) > 1 and segments[-1][-1] == len(curvature) - 1 and segments[0][0] == 0:
            if len(segments[0]) == len(segments[-1]):
                combined_segment = np.concatenate((segments[-1], segments[0]))
                # Remove the last segment
                segments.pop()
                # Replace the first segment with the combined segment
                segments[0] = combined_segment
                longest_segment = max(segments, key=len)
                try:
                  is_first_segment_longest = (longest_segment == segments[0]).all()
                  if is_first_segment_longest:
                    break
                except:
                  pass
          count += 1
        return θ_0_increment

    def getRequiredθ_0_increment(self):
        θ_0_increment = 0.1
        longest_segment = self.get_longest_straight_line()
        x_straight = self.curve_parametric()[0][longest_segment]
        y_straight = self.curve_parametric()[1][longest_segment]
        ret_x=ret_y = None
        if len(x_straight) == 0:
          return 0
        # Check if the line is nearly vertical because linear regression does not work so well with vertical lines
        # print(x_straight, np.abs(np.max(x_straight) - np.min(x_straight)), self.vertical_threshold)
        if np.abs(np.max(x_straight) - np.min(x_straight)) < self.vertical_threshold:  # Adjust threshold as needed
            x_value = np.mean(x_straight)  # You can choose mean, median, or another central value
            x_straight[:] = x_value
            ret_x = x_straight
            ret_y = y_straight
            self.slope = np.Infinity
            if y_straight[0] < y_straight[-1]:
              θ_0_increment = 90
            else:
              θ_0_increment = -90
        else:
            θ_0_increment = self.middle_of_line_at_0()

        return θ_0_increment
    def getdH(self): # get height
        self.setRequiredTheta0()
        longest_segment = self.get_longest_straight_line()
        x_straight = self.curve_parametric()[0][longest_segment]
        y_straight = self.curve_parametric()[1][longest_segment]
        if len(y_straight) == 0:
          self.dh = 0
          return 0
        dh = np.max(y_straight) - np.min(y_straight)
        self.dh = dh
        return dh
    def getdS(self):
        self.setRequiredTheta0()
        longest_segment = self.get_longest_straight_line()
        x_straight = self.curve_parametric()[0][longest_segment]
        if len(x_straight) == 0:
          self.ds = 0
          return 0
        ds = x_straight[-1] - x_straight[0] if x_straight[-1] > x_straight[0] else -x_straight[-1] + x_straight[0] if x_straight[-1] < x_straight[0] else 0
        # ds = x_straight[-1] - x_straight[0]
        self.ds = ds
        return ds
    def getStraightSectionFraction(self):
        self.setRequiredTheta0()
        longest_segment = self.get_longest_straight_line()
        x_straight = self.curve_parametric()[0][longest_segment]
        if len(x_straight) == 0:
          self.straightFraction = 0
          return 0
        fraction = (len(x_straight)/len(self.curve_parametric()[0]))*100
        fraction = round(fraction, 2)
        self.straightFraction = fraction
        return fraction
    def getmaxCurvature(self):
        self.setRequiredTheta0()
        _, _, curvature = self.curvature_parametric(self.R, self.f, self.a)
        maxCurvature = np.max(curvature)
        self.maxCurvature = maxCurvature
        return maxCurvature
    def plotallvalues(self, options):
        # "s":True, "h":True, "curvature":True, "fractional":True
        for key in options:
          if key == "s":
            self.plotSForWholeRange()
          if key == "h":
            self.plotHForWholeRange()
          if key == "curvatures":
            pass
          if key == "fractional":
            pass

        plt.tight_layout()
        pass
    def plotSForWholeRange(self):
      numGraphs = 4
      fig = axs = None
      if self.figS == None:
        fig, axs = plt.subplots(1, numGraphs, figsize=(12, 6))
        self.figS = fig
        self.axsS = axs
      else:
        fig = self.figS
        axs = self.axsS

      ### Plot for all R
      self.figS.suptitle(f"S for whole range (R:{self.R:.2f}, f:{self.f:.2f}, a:{self.a}, curvature:{self.curvature_threshold:.2f})")
      # Plot of R
      R_values = np.linspace(0, 10, 100)
      dS_values = []

      for R in R_values:
          # Update R value
          self.R = R
          # Calculate dS for the current R
          dS = self.getdS()
          dS_values.append(dS)
      for ax in axs:
          ax.clear()
      # Plot dS values against R values
      axs[0].plot(R_values, dS_values)
      axs[0].set_xlabel('R')
      axs[0].set_ylabel('S')
      axs[0].set_title('S vs. R')

      # plot of f
      f_values = np.linspace(0, 1, 100)
      dS_values = []

      for f in f_values:
          # Update R value
          self.f = f
          # Calculate dS for the current R
          dS = self.getdS()
          dS_values.append(dS)

      # Plot dS values against f values
      axs[1].plot(f_values, dS_values)
      axs[1].set_xlabel('f')
      axs[1].set_ylabel('S')
      axs[1].set_title('S vs. f')


      # plot of a
      a_values = np.linspace(-10, 10, 21)
      dS_values = []

      for a in a_values:
          # Update R value
          self.a = a
          # Calculate dS for the current R
          dS = self.getdS()
          dS_values.append(dS)

      # Plot dS values against R values
      axs[2].plot(a_values, dS_values)
      axs[2].set_xlabel('a')
      axs[2].set_ylabel('S')
      axs[2].set_title('S vs. a')

      ## curvature_threshold
      c_values = np.linspace(0, 3, 100)
      dS_values = []

      for c in c_values:
          # Update R value
          self.curvature_threshold = c
          # Calculate dS for the current R
          dS = self.getdS()
          dS_values.append(dS)

      # Plot dS values against R values
      axs[3].plot(c_values, dS_values)
      axs[3].set_xlabel('curvature_threshold')
      axs[3].set_ylabel('S')
      axs[3].set_title('S vs. curvature_threshold')

      plt.tight_layout()
    def plotHForWholeRange(self):
        numGraphs = 4
        fig = axs = None
        if self.figH == None:
            fig, axs = plt.subplots(1, numGraphs, figsize=(12, 6))
            self.figH = fig
            self.axsH = axs
        else:
            fig = self.figH
            axs = self.axsH

        ### Plot for all R
        self.figH.suptitle(f"H for whole range (R:{self.R:.2f}, f:{self.f:.2f}, a:{self.a}, curvature:{self.curvature_threshold:.2f})")
        # Plot of R
        R_values = np.linspace(0, 10, 100)
        dH_values = []

        for R in R_values:
            # Update R value
            self.R = R
            # Calculate dH for the current R
            dH = self.getdH()
            dH_values.append(dH)
        for ax in axs:
          ax.clear()
        # Plot dH values against R values
        axs[0].plot(R_values, dH_values)
        axs[0].set_xlabel('R')
        axs[0].set_ylabel('H')
        axs[0].set_title('H vs. R')

        # plot of f
        f_values = np.linspace(0, 1, 100)
        dH_values = []

        for f in f_values:
            # Update f value
            self.f = f
            # Calculate dH for the current f
            dH = self.getdH()
            dH_values.append(dH)

        # Plot dH values against f values
        axs[1].plot(f_values, dH_values)
        axs[1].set_xlabel('f')
        axs[1].set_ylabel('H')
        axs[1].set_title('H vs. f')


        # plot of a
        a_values = np.linspace(-10, 10, 21)
        dH_values = []

        for a in a_values:
            # Update a value
            self.a = a
            # Calculate dH for the current a
            dH = self.getdH()
            dH_values.append(dH)

        # Plot dH values against a values
        axs[2].plot(a_values, dH_values)
        axs[2].set_xlabel('a')
        axs[2].set_ylabel('H')
        axs[2].set_title('H vs. a')

        ## curvature_threshold
        c_values = np.linspace(0, 3, 100)
        dH_values = []

        for c in c_values:
            # Update curvature_threshold value
            self.curvature_threshold = c
            # Calculate dH for the current curvature_threshold
            dH = self.getdH()
            dH_values.append(dH)

        # Plot dH values against curvature_threshold values
        axs[3].plot(c_values, dH_values)
        axs[3].set_xlabel('curvature_threshold')
        axs[3].set_ylabel('H')
        axs[3].set_title('H vs. curvature_threshold')

        plt.tight_layout()
    def funcAnimation(self):
        numFrames = self.animationOptions["numFrames"]
        extra_keys, first_key = ([key for key in self.animationOptions.keys() if key != "numFrames"][1:],
                                   [key for key in self.animationOptions.keys() if key != "numFrames"][0])

        animationOptions={"numFrames":numFrames, first_key: self.animationOptions[first_key]}
        if self.plotType == "locus":
          # extra_keys, first_key = ([key for key in self.animationOptions.keys() if key != "numFrames"][1:],
          #                          [key for key in self.animationOptions.keys() if key != "numFrames"][0])
          ani = FuncAnimation(self.fig, self.updateAllPlot, frames=numFrames, interval=100, repeat=True, fargs=(numFrames, animationOptions, self.outputOptions))
          self.ani = ani.to_jshtml()
        elif self.plotType == "values":
          ani = None
          fig = None
          for key in self.outputOptions:
            if key == "s":
              fig = self.figS
            if key == "h":
              fig = self.figH
            if fig is not None:
              ani1 = FuncAnimation(fig, self.updateAllPlotForValues, frames=numFrames, interval=100, repeat=True, fargs=(numFrames, animationOptions, self.outputOptions))
              ani1 = ani1.to_jshtml()
              if ani is not None:
                ani = ani + ani1
              else:
                ani = ani1
          self.ani = ani

        _vars = self.vars
        for key in extra_keys:
          animationOptions={"numFrames":numFrames, key: self.animationOptions[key]}
          _vars["animationOptions"] = animationOptions
          self.addAnimation(_vars)

    def addAnimation(self, vars=None):
        anime = RR(vars)
        self.animes.append(anime)

    def showAnimation(self): # Not working
        ani = self.getAnimations()
        HTML(ani)
    def getAnimations(self):
        ani = self.ani
        animation_objects = [ani]

        for animation in self.animes:
            animation_objects.append(animation.ani)

        # Concatenate all animation objects into a single animation object
        combined_animation = animation_objects[0] if animation_objects else None
        for anim in animation_objects[1:]:
            combined_animation += anim
        return combined_animation


class parallelogram:
    def __init__(self, vars=None):
      pass


import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation

from ....utils import bmutils

class mechanism25(bmutils):
    def __init__(self, settings):
        self.settings = settings
        self.fig, self.ax = plt.subplots(figsize=(10, 6))
        self.ani = None

        # Extract parameters from settings
        self.R1 = self.settings['R1']
        self.R2 = self.settings['R2']
        self.R3 = self.settings['R3']
        self.R4 = self.settings['R4']
        self.R5 = self.settings['R5']
        self.theta_forward = self.settings['theta_forward']
        self.previous_theta = 0
        self.instances = settings["instances"]
        self.theta_limits = settings["theta_limits"]
        self.perimeter_theta = settings["perimeter_theta"]
        self.setup_plot()

    def setup_plot(self):
        self.ax.set_title(f'Mechanism 25')
        self.ax.set_xlim(-self.R5*2, self.R5*2)
        self.ax.set_ylim(-self.R5*2, self.R5*2)
        self.ax.axhline(0, color='black', linewidth=0.5, linestyle='--')
        # self.ax.grid()
        self.ax.axis("off")
        self.ax.set_aspect('equal')
        super().watermark(self.ax)
    
    def plot(self, theta = 0):    
        self.setup_plot()
        # the racks (fixed internal gears)
        x5, y5 = self.create_circle_section(self.R5, end_angle = 2*np.pi - self.theta_forward, phase=-self.theta_forward/2)
        x4, y4 = self.create_circle_section(self.R4, end_angle = self.theta_forward, phase=-np.pi/2-self.theta_forward/2)
        self.ax.plot(x5, y5, color='black', linestyle='--')
        self.ax.plot(x4, y4, color='blue', linestyle='--')


        x1, y1 = self.create_circle_section(self.R1, phase=theta)
        self.ax.plot(x1, y1, color='red', linestyle='--')
        self.ax.plot([0], [0], 'ro') 
        x1, y1 = self.create_circle_section(self.R1, phase=theta, end_angle=0.2*np.pi) # marker (solid section)
        self.ax.plot(x1, y1, color='red', linestyle='-', linewidth=5)

        d_theta_1 = theta - self.previous_theta
        self.previous_theta = theta
        R1 = self.R1
        R2 = self.R2
        R3 = self.R3
        R4 = self.R4
        R5 = self.R5
        d_theta_2 = R1/R2 * d_theta_1
        d_theta_3 = d_theta_2
        d_theta_5 = R1/R2 * R3/R5 * d_theta_1
        d_theta_4 = R1/R4 * d_theta_1

        # R2_center = [0, -R1-R2]

        ## rotate wheel about
        for index, instance in enumerate(self.instances):
            self.__dict__.update(instance)
            self.theta += d_theta_1
            R2_center = [0, -R1-R2]
            
            instance_theta = self.theta #+ d_theta_1
            if instance_theta != 0:
                direction = instance_theta/abs(instance_theta)
                self.previous_direction = direction
            else:
                direction = self.previous_direction
            self.previous_direction = direction
            instance_theta = direction * (abs(instance_theta) % self.perimeter_theta)
            if instance_theta < 0 :
                instance_theta += self.perimeter_theta

            d_theta_2 = - R1/R2 * instance_theta
            # check the section where instance_theta is
            if 0 <= instance_theta and instance_theta < self.theta_limits[0]: # first section
                d_theta_4 = instance_theta * (R1/R2) * (R3/R4)
                # d_theta_4 = R1/R4 * instance_theta
                centerx, centery = self.rotate_points(d_theta_4, [R2_center[0]],[R2_center[1]], [0,0])
                R2_center = [centerx[0],centery[0]]
            elif self.theta_limits[0] <= instance_theta and instance_theta < self.theta_limits[1]:
                
                d_theta_5 = self.theta_forward/2 + R1/R5 * (instance_theta-self.theta_limits[0])
                centerx, centery = self.rotate_points(d_theta_5, [R2_center[0]],[R2_center[1]], [0,0])
                R2_center = [centerx[0],centery[0]]
            elif self.theta_limits[1] <= instance_theta and instance_theta < self.theta_limits[2]:
                # d_theta_4 = (2*np.pi - self.theta_forward/2) + R1/R4 * (instance_theta-self.theta_limits[1]) 
                d_theta_4 = (2*np.pi - self.theta_forward/2) + (instance_theta-self.theta_limits[1]) * (R1/R2) * (R3/R4)
                centerx, centery = self.rotate_points(d_theta_4, [R2_center[0]],[R2_center[1]], [0,0])
                R2_center = [centerx[0],centery[0]]
                pass

            x2, y2 = self.create_circle_section(self.R2, phase=d_theta_2)
            x3, y3 = self.create_circle_section(self.R3, phase=d_theta_2)
            # if index == 0: 
            self.ax.plot(x2+R2_center[0], y2+R2_center[1], color='blue', linestyle='--')
            self.ax.plot(x3+R2_center[0], y3+R2_center[1], color='black', linestyle='--')

            x2, y2 = self.create_circle_section(self.R2, phase=d_theta_2, end_angle=0.2*np.pi) # marker (solid section)
            x3, y3 = self.create_circle_section(self.R3, phase=d_theta_2, end_angle=0.05*np.pi) # marker (solid section)
            self.ax.plot(x2+R2_center[0], y2+R2_center[1], color='blue', linestyle='-', linewidth=5)
            self.ax.plot(x3+R2_center[0], y3+R2_center[1], color='black', linestyle='-', linewidth=5)
            # if index == 1: 
            #     self.ax.plot(x2+R2_center[0], y2+R2_center[1], color='blue', linestyle='--')
            #     self.ax.plot(x3+R2_center[0], y3+R2_center[1], color='green', linestyle='--')
            
            for key, value in self.__dict__.items():
                if key in instance:  # Check if the key exists in the instance
                    instance[key] = value  # Update the dictionary
                elif hasattr(instance, '__dict__') and key in vars(instance):
                    setattr(instance, key, value)
            pass

        return self
    
    def update(self, interval=100, frames=100):
        speed = self.settings["speed"]
        len = self.settings["animation_rounds"]
        displacements = np.linspace(0, len, int(250/speed))
        displacements = np.concatenate([displacements, displacements[::-1]])  # Reverse motion
        def animate(displacement):
            self.ax.clear()
            self.plot(displacement)

        
        self.setup_plot()
        self.ani = animation.FuncAnimation(self.fig, lambda displacement: animate(displacement), frames=displacements, interval=50, repeat=True)
        return self
        

    @staticmethod
    # theta_forward in radians
    # theta(offset) in perimeter fraction
    def Settings(R1=5,R2=4, theta_forward=0.5*np.pi, return_fraction=1/3, animation_rounds=2, speed=0.5, num_instances=2):
        w_forward = theta_forward/(1 - return_fraction)
        w_return = (2*np.pi-theta_forward)/(return_fraction)

        # R1 radius of sun
        # R2 Radius of smaller of the 2 in compound gear
        # R3 Radius of bigger of the 2 in compound gear
        k = w_return/w_forward
        R3 = R2 / k
        R4 = R1 + R2 + R3
        R5 = R1 + 2 * R2

        theta_limits = [] # theta for R1 to complete the various sections
        theta_1_section_1 = theta_forward/2 * (R2/R1) * (R4/R3)
        # print((R1/R2) * (R3/R4))

        theta_limits.append(theta_1_section_1) # OK
        theta_return = 2*np.pi - theta_forward
        s_limit_2_R5 = theta_return * R5
        # theta_3 = s_limit_2_R5/ R2

        theta_1_section_2 = s_limit_2_R5/R1 # theta_2 = theta_3
        theta_limits.append(theta_1_section_2+theta_1_section_1)
        theta_limits.append(theta_1_section_1+theta_1_section_2+theta_1_section_1)
        perimeter_theta = theta_limits[-1]
        animation_rounds = int(animation_rounds * perimeter_theta)
        theta_points = [(perimeter_theta/num_instances) * i for i in range(num_instances)]
        instances = [{"theta":theta, "rim_theta":((2*np.pi/num_instances )*index)} for index, theta in enumerate(theta_points)]
        for instance in instances:
            instance["previous_direction"] = 1
        # instances[0]["theta"] = perimeter_theta/2
        
        
        return {
            "R1": R1,
            "R2": R2,
            "R3": R3,
            "R4": R4,
            "R5": R5,
            "theta_forward":theta_forward,
            "animation_rounds":animation_rounds,
            "speed":speed,
            "perimeter_theta":perimeter_theta,
            "theta_limits":theta_limits,
            "instances":instances,
        }
    @staticmethod
    def info():
        description = (
            "Mechanism Description:\n"
            "-----------------------\n"
            "Continuous rotary motion with different constant velocities for different sections.\n"
            "**Applications**:\n"
            "This type of mechanism is intended to be used in the development of clamp-on self-actuated long-range linear motion mechanisms. and 2-spoke rimless wheels\n"
        )
        return description


import numpy as np
import matplotlib.pyplot as plt
import math
from matplotlib.animation import FuncAnimation, PillowWriter


class Settings():
    def __init__(self, radius, center_distance, leg_height, leg_length, num_legs, fig=None, ax=None) -> None:
        self.position = 0
        self.radius = radius 
        self.center_distance = center_distance
        self.leg_height = leg_height # vertical
        self.leg_length = leg_length # horizontal
        self.num_legs = num_legs

        self.points_on_curve = []

        self.endPerimeter = endPerimeter = np.pi * radius
        self.chainPerimeter = 2 * (self.endPerimeter + center_distance)

        self.legPositions = [self.chainPerimeter/self.num_legs * i for i in range(self.num_legs)]
        self.legPositionsStop = [pos + self.leg_length for pos in self.legPositions]
        self.head_starts = [pos for pos in self.legPositions]
        self.full_leg_height = self.radius + self.leg_height

        self.checkPoints = [
            center_distance / 2,  # Quadrant 1
            center_distance / 2 + endPerimeter,  # Quadrant -1
            center_distance / 2 + endPerimeter + center_distance / 2,  # Quadrant 2
            center_distance / 2 + endPerimeter + center_distance,  # Quadrant 3
            center_distance / 2 + endPerimeter + center_distance + endPerimeter,  # Quadrant -2
            center_distance / 2 + endPerimeter + center_distance + endPerimeter + center_distance / 2  # Quadrant 4
        ]

        self.fig = fig
        self.ax = ax

        if ax is not None:
            ax.set_xlim(-1.5*self.full_leg_height-0.5*center_distance, 1.5*self.full_leg_height+0.5*center_distance)
            height = 1.5*self.full_leg_height+radius
            ax.set_ylim(-height, height)
            ax.set_aspect('equal')

            fig.suptitle('Legged Tracks', fontsize=14)
            ax.set_title(f"r={radius}, d={center_distance}, n={num_legs}, l={leg_length}", fontsize=10)
            # ax.axis('off')
            ax.grid(True)

        colours = [
            ['k', 'r', 'g', 'b'],  # Set 1
            ['c', 'm', 'y', 'k'],  # Set 2
            ['m', 'y', 'b', 'g'],  # Set 3
            ['r', 'g', 'b', 'c'],  # Set 4
            ['y', 'b', 'g', 'k'],  # Set 5
            ['k', 'c', 'r', 'm'],  # Set 6
            ['g', 'y', 'c', 'b'],  # Set 7
            ['b', 'k', 'y', 'r'],  # Set 8
            ['m', 'c', 'k', 'g'],  # Set 9
            ['r', 'y', 'm', 'b']   # Set 10
        ]
        
        # Assign colors based on num
        self.colours = [colours[i % 10] for i in range(num_legs)] 
        

        pass


### This class has a wrong algorith which will give wrong results.
class leggedTrack():
    def __init__(self,settings) -> None:
        for attr, value in vars(settings).items():
            setattr(self, attr, value)
        
        self.createPaths()
        self.createLegPoints()
        self.ani = None
        # self.points_on_curve = [] For some reason this only works when defined in settings and not here

    def createPaths(self):
        self.createChain(self.radius, "-", curvePoints=True)
        self.createChain(self.full_leg_height, "--", False)

    def createChain(self, radius, dashed, gearDots = True, curvePoints= False):

        # Define the points for the horizontal sides
        x1 = -self.center_distance / 2
        x2 = self.center_distance / 2
        y = radius

        ax = self.ax

        center_distance = self.center_distance

        # Draw the flat horizontal sides
        ax.plot([x1, x2], [y, y], f'k{dashed}')  # Top side
        ax.plot([x1, x2], [-y, -y], f'k{dashed}')  # Bottom side
        if curvePoints:
            num_points = 10000
        else:
            num_points = 100

        # Draw the semicircles
        theta_left = np.linspace(0.5*np.pi, 1.5* np.pi, num_points)
        theta_right = np.linspace(-0.5*np.pi, 0.5* np.pi, num_points)
        semicircle_left = [(-self.center_distance / 2 + radius * np.cos(theta_left)),radius * np.sin(theta_left)]
        semicircle_right_side_of_left = [(-self.center_distance / 2 + radius * np.cos(theta_right)),radius * np.sin(theta_right)]
        semicircle_right = [(self.center_distance / 2 + radius * np.cos(theta_right)),radius * np.sin(theta_right)]
        semicircle_left_side_of_right = [(self.center_distance / 2 + radius * np.cos(theta_left)),radius * np.sin(theta_left)]

        ax.plot(semicircle_left[0], semicircle_left[1], f'k{dashed}')
        ax.plot(semicircle_right[0], semicircle_right[1], f'k{dashed}')
        if curvePoints:
            ax.plot(semicircle_right_side_of_left[0], semicircle_right_side_of_left[1], f'k{dashed}')
            ax.plot(semicircle_left_side_of_right[0], semicircle_left_side_of_right[1], f'k{dashed}')

        # Add dots at the centers of the semicircles
        center_top_left = (-center_distance / 2, 0)
        center_top_right = (center_distance / 2, 0)

        ax.plot(center_top_left[0], center_top_left[1], 'ko')  # Center of left semicircle
        ax.plot(center_top_right[0], center_top_right[1], 'ko')  # Center of right semicircle

        if gearDots:
            # Add small dots at the circumference of the semicircles
            self.gear_dot_left, = ax.plot(center_top_left[0] + radius, 0, 'ro')  # Starting position for left semicircle
            self.gear_dot_right, = ax.plot(center_top_right[0] + radius, 0, 'ro')  # Starting position for right semicircle

            # Add lines from the center to the circumference dots
            self.gear_line_top_left, = ax.plot([center_top_left[0], center_top_left[0] + radius], [center_top_left[1], 0], 'r-')
            self.gear_line_top_right, = ax.plot([center_top_right[0], center_top_right[0] + radius], [center_top_right[1], 0], 'r-')



         # Define the number of points along each horizontal line

        # Create points for horizontal lines
        x_horizontal = np.linspace(x1, x2, num_points)
        y_top = np.full_like(x_horizontal, y)
        y_bottom = np.full_like(x_horizontal, -y)

        if curvePoints:
            self.points_on_curve = (
                list(zip(semicircle_left[0], semicircle_left[1])) +
                list(zip(semicircle_right[0], semicircle_right[1])) +
                list(zip(x_horizontal, y_top)) +
                list(zip(x_horizontal, y_bottom)) +
                [(x1, y), (x2, y), (x1, -y), (x2, -y)]
            )
            positive_y_points = [(x, y) for x, y in self.points_on_curve if y >= 0]
            negative_y_points = [(x, y) for x, y in self.points_on_curve if y < 0]

            # Sort positive y points:
            # 1. First take values for x from 0 to the least x
            first_positive_y_points = sorted([p for p in positive_y_points if p[0] <= 0], key=lambda p: p[0], reverse=True)
            # 2. Then take values for x from greatest x to 0
            second_positive_y_points = sorted([p for p in positive_y_points if p[0] > 0], key=lambda p: p[0], reverse=True)

            # Sort negative y points from least x to greatest x
            sorted_negative_y_points = sorted(negative_y_points, key=lambda p: p[0])

            # Combine all points in the desired order
            sorted_points = first_positive_y_points + sorted_negative_y_points + second_positive_y_points 

            # Update points on curve
            self.points_on_curve = sorted_points
    def find_leg_theta(self, current_point, L, debug=False):
        x, y = current_point
        start_index = None
        min_distance = float('inf')
        
        # Find the index of the point on the curve that maps to the current point
        # for i, (px, py) in enumerate(self.points_on_curve):
        #     if np.isclose(px, x, atol=1e-3) and np.isclose(py, y, atol=1e-3):
        #         start_index = i
        #         break

        # Find the index of the closest point on the curve to the current point
        for i, (px, py) in enumerate(self.points_on_curve):
            distance = np.hypot(px - x, py - y)
            
            if distance < min_distance:
                if debug:
                    print("is less:", distance, min_distance)
                # print(distance)
                min_distance = distance
                start_index = i

        
        if start_index is None:
            if debug:
                print("is None", x,y, len(self.points_on_curve))
            return None, None
        

        # Search for points after the current point index
        for i in range(start_index + 1, len(self.points_on_curve)):
            px, py = self.points_on_curve[i]
            dx = px - x
            dy = py - y
            distance = np.hypot(dx, dy)
            if np.isclose(distance, L, atol=1e-3):
                theta = np.arctan2(dy, dx)
                return theta, (px, py)

        # Wrap around and search from the beginning up to the current point index
        for i in range(0, start_index):
            px, py = self.points_on_curve[i]
            dx = px - x
            dy = py - y
            distance = np.hypot(dx, dy)
            if np.isclose(distance, L, atol=1e-3):
                theta = np.arctan2(dy, dx)
                return theta, (px, py)

        return None, None


    def setPosition(self, pointIndex, addToPosition = 0):
        head_start = self.head_starts[pointIndex] + addToPosition
        position = self.position
        checkPoints = self.checkPoints
        chainPerimeter  = self.chainPerimeter
        center_distance = self.center_distance
        position+=head_start
        # Calculate the quadrant and remaining position

        remPosition = position % chainPerimeter
        position = remPosition
        for i, point in enumerate(checkPoints):
            if remPosition < point:
                rackQuadrant = i
                break

        armAngle = 0.5*np.pi  # degrees

        if rackQuadrant == 0:  # [0]
            pointPosition = -position  # a negative value
        elif rackQuadrant == 1:  # [1] # Dwell
            armAngle = np.radians(270 - 180 * (checkPoints[1] - position) / (checkPoints[1] - checkPoints[0]))
            pointPosition = -(center_distance / 2)  # negative value
        elif rackQuadrant == 2:  # [2]
            armAngle = 1.5*np.pi
            pointPosition = -(checkPoints[2] - position)  # negative value
        elif rackQuadrant == 3:  # [0]
            armAngle = 1.5*np.pi
            pointPosition = (center_distance / 2) - (checkPoints[3] - position)  # +ve value
        elif rackQuadrant == 4:  # Dwell
            armAngle = np.radians((360+90) - 180 * (checkPoints[4] - position) / (checkPoints[4] - checkPoints[3]))
            pointPosition = center_distance / 2
        else:  # rackQuadrant == 5
            pointPosition = checkPoints[5] - position  # +ve value

        
        x = pointPosition +  self.radius * np.cos(armAngle)
        y = self.radius * np.sin(armAngle)

        return pointPosition, armAngle, [x,y]
    
    def get_line_angle(self, x1, y1, x2, y2):
        # Calculate the angle in radians
        angle_rad = math.atan2(y2 - y1, x2 - x1)
        
        # # Convert the angle from radians to degrees
        # angle_deg = math.degrees(angle_rad)
        
        return angle_rad#, angle_deg
    
    def createLegPoints(self):
        self.legPoints = []
        self.legEndPoints = []
        self.legBaseLines = []
        self.legVerticalLine1s = [ ]
        self.legVerticalLine2s = [ ]
        self.legFootPoints = [ ]
        try:
            self.position=1
            for index in range(len(self.head_starts)):
                _,armAngle,pos = self.setPosition(index)
                x = pos[0]
                y = pos[1]
                point, = self.ax.plot([x],[y], f"{self.colours[index][0]}o")
                self.legPoints.append(point)


                leg_theta,pos = self.find_leg_theta(pos, self.leg_length)
                # print(leg_theta)
                # _,_,pos = self.setPosition(index, self.leg_length)
                x1 = pos[0]
                y1 = pos[1]
                # x1 = x + self.leg_length * np.cos(leg_theta)
                # y1 = y + self.leg_length * np.sin(leg_theta)
                pointEnd, = self.ax.plot([x1],[y1], f"{self.colours[index][0]}o")
                self.legEndPoints.append(pointEnd)

                armAngle = self.get_line_angle(x,y,x1,y1) - 0.5 * np.pi

                vLine1EndPoint = [x + self.leg_height * np.cos(armAngle), y + self.leg_height * np.sin(armAngle) ]
                vLine2EndPoint = [x1 + self.leg_height * np.cos(armAngle), y1 + self.leg_height * np.sin(armAngle) ]

                # a line connecting point and pointEnd
                line, = self.ax.plot([x,x1],
                                 [y,y1],
                                 color=self.colours[index][0])
                vLine1, = self.ax.plot([x,vLine1EndPoint[0]],
                                 [y,vLine1EndPoint[1]],
                                 color=self.colours[index][0])
                vLine2, = self.ax.plot([x1,vLine2EndPoint[0]],
                                 [y1,vLine2EndPoint[1]],
                                 color=self.colours[index][0], linestyle='--')
                hLine, = self.ax.plot([vLine1EndPoint[0],vLine2EndPoint[0]],
                                 [vLine1EndPoint[1],vLine2EndPoint[1]],
                                 color=self.colours[index][0])
                
                self.legBaseLines.append(line)
                self.legVerticalLine1s.append(vLine1)
                self.legVerticalLine2s.append(vLine2)
                self.legFootPoints.append(hLine)

        except IndexError as e:
            print(f"Error: {e}. Ensure that all indices in self.head_starts are within the range of self.setPosition.")
            return

    def show(self):
        plt.show()
        
    def save(self, fileName):
        if self.ani is not None:
            writer = PillowWriter(fps=25)
            self.ani.save(f"{fileName}.gif", writer=writer)
        else:
            plt.savefig(f"{fileName}.png")
        return self

    def update_plots(self, position):
        self.position = position
        radius = self.radius
        center_distance = self.center_distance
        angle = self.position/radius
        center_top_left = (-center_distance / 2, 0)
        center_top_right = (center_distance / 2, 0)
        # Update the position of the dots
        self.gear_dot_left.set_data([center_top_left[0] + radius * np.cos(angle)], [radius * np.sin(angle)])
        self.gear_dot_right.set_data([center_top_right[0] + radius * np.cos(angle)], [radius * np.sin(angle)])
        # Update the position of the lines
        self.gear_line_top_left.set_data([center_top_left[0], center_top_left[0] + radius * np.cos(angle)], 
                               [center_top_left[1], radius * np.sin(angle)])
        self.gear_line_top_right.set_data([center_top_right[0], center_top_right[0] + radius * np.cos(angle)], 
                                [center_top_right[1], radius * np.sin(angle)])
        # return self.gear_dot_left,  self.gear_dot_right, self.gear_line_top_left, self.gear_line_top_right
        # Update leg points and lines
        for index in range(len(self.head_starts)):
            _, _, pos = self.setPosition(index)
            x = pos[0]
            y = pos[1]
            self.legPoints[index].set_data([x], [y])

            #  _,armAngle,pos = self.setPosition(index)
            #     x = pos[0]
            #     y = pos[1]

            # _, _, pos = self.setPosition(index, self.leg_length)
            leg_theta,pos = self.find_leg_theta(pos, self.leg_length, debug=False)
            if pos is None:
                continue
            x1 = pos[0]
            y1 = pos[1]
            
            self.legEndPoints[index].set_data([x1], [y1])

            self.legBaseLines[index].set_data([x, x1], [y, y1])

            armAngle = self.get_line_angle(x, y, x1, y1) - 0.5 * np.pi

            vLine1EndPoint = [x + self.leg_height * np.cos(armAngle), y + self.leg_height * np.sin(armAngle)]
            vLine2EndPoint = [x1 + self.leg_height * np.cos(armAngle), y1 + self.leg_height * np.sin(armAngle)]

            # # Update vertical lines
            self.legVerticalLine1s[index].set_data([x, vLine1EndPoint[0]], [y, vLine1EndPoint[1]])
            self.legVerticalLine2s[index].set_data([x1, vLine2EndPoint[0]], [y1, vLine2EndPoint[1]])

            # # Update horizontal line
            self.legFootPoints[index].set_data([vLine1EndPoint[0], vLine2EndPoint[0]], [vLine1EndPoint[1], vLine2EndPoint[1]])

        return (self.gear_dot_left, self.gear_dot_right,
                self.gear_line_top_left, self.gear_line_top_right
                 ,*self.legBaseLines,
                 *self.legPoints, *self.legEndPoints,
                *self.legVerticalLine1s, *self.legVerticalLine2s,
                *self.legFootPoints
                )

    def update(self, stopPosition, speed):
        if self.ani is not None:
            self.ani.event_source.stop()  # Stop previous animation if exists
        stopPosition = int(stopPosition)
        frames_forward = np.arange(0, stopPosition+1,speed)
        frames_backward = np.arange(stopPosition, -1, -speed)
        frames = np.concatenate((frames_forward, frames_backward))

        self.ani = FuncAnimation(self.fig, self.update_plots, frames=frames, blit=True, interval=100)
        return self



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from matplotlib.animation import FuncAnimation, PillowWriter


class mechanism3:
    def __init__(self, settings):
        self.settings = settings
        self.ani = None
        pass 

    @staticmethod
    def calculate_radius2(radius1, gear1_angle):
        radius2 = (gear1_angle * radius1) / (2 * np.pi - gear1_angle)
        return radius2
    
    @staticmethod
    def Settings(options, center_distance = 0, fig = None, ax = None, grid=False, rotations = 1, speed = 0.5, n=3):
        """
        Populate the settings dictionary with values from the options dictionary.

        Parameters:
        options (dict): A dictionary containing options and their values.

        Returns:
        dict: A dictionary containing the populated settings.
        """
        
        displacements = []
        radii_2 = []
        num_instances = len(options)
        origins = []
        phase_diff = 2*np.pi/num_instances
        
        for index, instance in enumerate(options):
            # if instance.get('offset_angle_degrees'):
            #     instance["offset_angle"] = np.radians(instance.get('offset_angle_degrees'))
            n = instance.get('n',n)
            instance["n"] = n
            speed = instance.get('speed',speed)
            instance["speed"] = speed
            segment_angle = 2*np.pi / n
            instance["segment_angle"] = segment_angle
            # instance["offset_angle"] = (-segment_angle/2-np.pi/2 + 2*np.pi) % (2*np.pi)
            # instance["offset_angle"] = (-segment_angle + 2*np.pi) % (2*np.pi)
            instance["offset_angle"] = (-np.pi/2 + 2*np.pi) % (2*np.pi)
            instance["start_angle_1"] = instance["offset_angle"]
            instance["stop_angle_1"] = instance.get('start_angle_1') + segment_angle

            # instance["radius2"] = mechanism3.calculate_radius2(instance.get('radius1', 0), forward_angle)
            radii_2.append(instance["radius1"]) ## fix
            displacements.append(segment_angle * instance.get('radius1', 0))
            instance["displacement"] = segment_angle * instance.get('radius1', 0)
            instance["x_limit"] = instance["displacement"]+10
            if center_distance == 0:
                center_distance =2*segment_angle * instance.get('radius1', 0)+10
            origins.append(index * center_distance)

            

        settings = {"instances": options}
        if fig:
            settings['fig'] = fig 
        else:
            fig, ax = plt.subplots()
            ax.set_aspect('equal')
        settings['fig'] = fig 
        settings['ax'] = ax 

        # x_limit = sum(displacements) + (len(displacements) - 1) * center_distance
        x_limit = (len(options)) * center_distance + 10
        y_limit = 2 * max(radii_2) *5

        settings["x_limit"] = x_limit
        settings["y_limit"] = y_limit


        ax.set_xlim(-x_limit/2, x_limit/2)
        ax.set_ylim(-y_limit/2, y_limit/2)
        ax.set_aspect('equal')

        if grid:
            plt.grid(True)

        if origins:
            mid_value = (max(origins) + min(origins)) / 2
            origins = [o - mid_value for o in origins]

        # Assign origins to each instance
        for index, instance in enumerate(options):
            instance["origin"] = origins[index]

        settings["rotations"] = rotations
        settings["speed"] = speed

        return settings

    def create_circle_section(self, r, num_points=100, start_angle=0, end_angle=2 * np.pi, phase=0):
        theta = np.linspace(start_angle+phase, end_angle+phase, num_points)
        x = r * np.cos(theta)
        y = r * np.sin(theta)
        return x, y

    def rack_displacement(self, angle, radius1, segment_angle):
        # angle = angle %(2*np.pi)
        # reversing_angle = angle
        # if angle <= segment_angle:
        #     displacement = -angle * radius1
            
        # elif angle <= 2 * segment_angle:
        #     last_position = -segment_angle * radius1
        #     displacement = last_position + (angle - segment_angle) * radius1 
        #     reversing_angle *= -1
        # elif angle <= 3 * segment_angle:
        #     last_position = 0
        #     displacement = last_position - (angle - 2*segment_angle) * radius1 
        # else:
        #     last_position = -segment_angle * radius1
        #     displacement = last_position + (angle - 3*segment_angle) * radius1
        #     reversing_angle *= -1
        angle = angle % (2 * np.pi)  # Normalize angle
        reversing_angle = angle
        displacement = 0
        last_position = 0

        # Iterate through segments
        num_segments = int(np.ceil((2 * np.pi) / segment_angle))
        for i in range(num_segments):
            # Calculate the start and end angle of the current segment
            start_angle = i * segment_angle
            end_angle = (i + 1) * segment_angle
            # Update last_position for the next segment
            if i % 2 == 0:
                last_position = 0 #start_angle * radius1
            else:
                last_position = -segment_angle * radius1 #start_angle * radius1 - segment_angle * radius1

            if angle <= end_angle:
                # Calculate displacement for the current segment
                if i % 2 == 0:
                    # displacement = last_position - (angle - 2*segment_angle) * radius1
                    displacement = last_position - (angle - start_angle) * radius1
                else:
                    displacement = last_position + (angle - start_angle) * radius1
                    reversing_angle = -1 * angle

                break

            
        
        displacement *= -1
        return (displacement,reversing_angle)

    def plot(self, instance_number=None):
        fig = self.settings['fig']
        ax = self.settings['ax']

        if instance_number is not None:
            # Filter instances to only include the specified instance_number
            instances = [self.settings['instances'][instance_number]]
        else:
            # Plot all instances
            instances = self.settings['instances']

        for inst in instances:
            radius1 = inst.get('radius1', 0)
            radius2 = inst.get('radius2', 0)
            start_angle_1 = inst.get('start_angle_1', 0)
            stop_angle_1 = inst.get('stop_angle_1', 0)
            origin = inst.get('origin', 0)
            x_limit = inst.get("x_limit")
            offset_angle = inst.get("offset_angle")
            segment_angle = inst.get("segment_angle")

            section_colors = ["blue", "red"]
            # for i in range(self.settings["n"]):
            for i in range(inst.get("n")):
                color_index = i%len(section_colors)
                x1, y1 = self.create_circle_section(radius1, start_angle=start_angle_1+i*segment_angle, end_angle=stop_angle_1+i*segment_angle)
                ax.plot(x1 + origin, y1, label=f'Larger Circle Section (Origin {origin})', color=section_colors[color_index], linestyle='--')

                stop_angle = (stop_angle_1 +i*segment_angle)%(2*np.pi)
                start_angle = (start_angle_1 +i*segment_angle)%(2*np.pi)
                x_end_circle1 = radius1 * np.cos(stop_angle)
                y_end_circle1 = radius1 * np.sin(stop_angle)
                x_start_circle1 = radius1 * np.cos(start_angle)
                y_start_circle1 = radius1 * np.sin(start_angle)

                ax.plot([0+ origin, x_end_circle1+ origin], [0, y_end_circle1], color='black', linestyle='-')
            ax.plot(origin, 0, 'ko')

            # pinion 2
            x1, y1 = self.create_circle_section(radius2, start_angle=0, end_angle=2*np.pi)
            ax.plot(x1 + origin, y1-radius1-radius2, color="blue", linestyle='--')
            ax.plot(origin, -radius1-radius2, 'ko')

            # # # Add dashed horizontal lines
            rack_displacement = 0
            ax.plot([-x_limit/2 + origin+rack_displacement, x_limit/2 + origin+rack_displacement], [-radius1, -radius1], color='red', linestyle='--')
            ax.plot([-x_limit/2 + origin+rack_displacement, x_limit/2 + origin+rack_displacement], [-radius1-radius2*2, -radius1-radius2*2], color='blue', linestyle='--')
            ax.plot([-x_limit/2 + origin+rack_displacement, x_limit/2 + origin+rack_displacement], [-radius1-radius2*4, -radius1-radius2*4], color='black', linestyle='--')

            # # # Connect the starts and ends of the lines
            ax.plot([-x_limit/2 + origin+rack_displacement, -x_limit/2 + origin+rack_displacement], [-radius1, -radius1-radius2*4], color='black', linestyle='-')
            ax.plot([x_limit/2+origin+rack_displacement, x_limit/2+origin+rack_displacement], [-radius1, -radius1-radius2*4], color='black', linestyle='-')

            x1, y1 = self.create_circle_section(radius1, start_angle=0, end_angle=2*np.pi)
            ax.plot(x1 + origin, y1-radius1*2-radius2*4, color="black", linestyle='--')
            ax.plot(origin, -radius1*2-radius2*4, 'ko')

            ax.plot([0+ origin, 0+ origin], [-radius1*2-radius2*4, -radius1*3-radius2*4], color='black', linestyle='-')


        # Set plot settings
        ax.set_aspect('equal', adjustable='box')
        ax.set_title(f'Mechanism 3')
        ax.set_xlabel('X')
        ax.set_ylabel('Y')
        ax.grid(True)
        ax.axis('off')
        
        

        return self


    def show(self):
        plt.show()
        return self
    
    def save(self, fileName):
        if self.ani is not None:
            writer = PillowWriter(fps=25)
            self.ani.save(f"{fileName}.gif", writer=writer)
        else:
            plt.savefig(f"{fileName}.png")
        return self
    
    def update(self, interval=100, frames=100):
        fig = self.settings['fig']
        ax = self.settings['ax']

        rotation = self.settings.get('rotation', 2 * np.pi)  # Default to 360 degrees (full rotation)
        speed = self.settings.get('speed', 1)  # Default speed factor

        # Calculate the total number of frames needed for the given rotation
        total_frames = int((rotation / (2 * np.pi)) * frames / speed)
        
        def animate(frame):
            ax.clear()  # Clear the current plot
            for inst in self.settings['instances']:
                phase = (frame * rotation / total_frames)

                radius1 = inst.get('radius1', 0)
                radius2 = inst.get('radius2', 0)
                start_angle_1 = inst.get('start_angle_1', 0)
                stop_angle_1 = inst.get('stop_angle_1', 0)
                origin = inst.get('origin', 0)
                x_limit = inst.get("x_limit")
                offset_angle = inst.get("offset_angle")
                segment_angle = inst.get("segment_angle")

                section_colors = ["blue", "red"]
                for i in range(inst["n"]):
                    color_index = i%len(section_colors)
                    x1, y1 = self.create_circle_section(radius1, start_angle=start_angle_1+i*segment_angle, end_angle=stop_angle_1+i*segment_angle, phase=phase)
                    ax.plot(x1 + origin, y1, label=f'Larger Circle Section (Origin {origin})', color=section_colors[color_index], linestyle='--')

                    stop_angle = (stop_angle_1 +i*segment_angle+phase)%(2*np.pi)
                    start_angle = (start_angle_1 +i*segment_angle+phase)%(2*np.pi)
                    x_end_circle1 = radius1 * np.cos(stop_angle)
                    y_end_circle1 = radius1 * np.sin(stop_angle)
                    x_start_circle1 = radius1 * np.cos(start_angle)
                    y_start_circle1 = radius1 * np.sin(start_angle)

                    ax.plot([0+ origin, x_end_circle1+ origin], [0, y_end_circle1], color='black', linestyle='-')
                ax.plot(origin, 0, 'ko')

                rack_displacement,reverse_angle = self.rack_displacement(phase, radius1, inst.get("segment_angle"))
                # pinion 2
                x1, y1 = self.create_circle_section(radius2, start_angle=0, end_angle=2*np.pi,phase=reverse_angle*inst["n"])
                ax.plot(x1 + origin, y1-radius1-radius2, color="blue", linestyle='--')
                ax.plot(origin, -radius1-radius2, 'ko')

                # # # # Add dashed horizontal lines
                # rack_displacement = 0
                ax.plot([-x_limit/2 + origin+rack_displacement, x_limit/2 + origin+rack_displacement], [-radius1, -radius1], color='red', linestyle='--')
                ax.plot([-x_limit/2 + origin+rack_displacement, x_limit/2 + origin+rack_displacement], [-radius1-radius2*2, -radius1-radius2*2], color='blue', linestyle='--')
                ax.plot([-x_limit/2 + origin+rack_displacement, x_limit/2 + origin+rack_displacement], [-radius1-radius2*4, -radius1-radius2*4], color='black', linestyle='--')

                # # # Connect the starts and ends of the lines
                ax.plot([-x_limit/2 + origin+rack_displacement, -x_limit/2 + origin+rack_displacement], [-radius1, -radius1-radius2*4], color='black', linestyle='-')
                ax.plot([x_limit/2+origin+rack_displacement, x_limit/2+origin+rack_displacement], [-radius1, -radius1-radius2*4], color='black', linestyle='-')

                x1, y1 = self.create_circle_section(radius1, start_angle=0, end_angle=2*np.pi, phase=reverse_angle*-1)
                ax.plot(x1 + origin, y1-radius1*2-radius2*4, color="black", linestyle='--')
                ax.plot(origin, -radius1*2-radius2*4, 'ko')


                max_displacement,_ = self.rack_displacement(inst.get("segment_angle"), radius1, inst.get("segment_angle"))
                output_angle = -1*(rack_displacement/max_displacement * inst.get("segment_angle")) +1.5* np.pi
                # if output_angle > 0:
                #     output_angle = -output_angle
                x_end_circle1 = radius1 * np.cos((output_angle)%(2*np.pi))
                y_end_circle1 = radius1 * np.sin((output_angle)%(2*np.pi))
                ax.plot([0+ origin, x_end_circle1+ origin], [-radius1*2-radius2*4, y_end_circle1-radius1*2-radius2*4], color='black', linestyle='-')

            ax.set_aspect('equal', adjustable='box')
            ax.set_title(f'Mechanism 3')
            ax.set_xlabel('X')
            ax.set_ylabel('Y')
            ax.grid(True)
            x_limit = self.settings["x_limit"]
            y_limit = self.settings["y_limit"]
            self.settings["x_limit"] = x_limit
            self.settings["y_limit"] = y_limit


            ax.set_xlim(-x_limit/2, x_limit/2)
            ax.set_ylim(-y_limit/2, y_limit/2)
            ax.axis('off')

        self.ani = animation.FuncAnimation(fig, animate, frames=total_frames, interval=interval, repeat=True)
        return self
    
    @staticmethod
    def info():
        description = (
            "Mechanism Description:\n"
            "-----------------------\n"
            "This mechanism is a for converting constant velocity rotary input to oscillating rotary output. "
            "It is constructed using a compound mangled pinion, 2 full pinions and 2 racks:\n"
            "\n"
            "Use with { n ∈ ℤ ∣ n is even and n > 2 }"
            "\n"
            "- **Constant Velocity:** The mechanism ensures a uniform linear speed during the forward stroke, providing smooth and consistent motion.\n"
            "- **Quick Return:** It incorporates a quick return feature, which means the return stroke (or reverse movement) is faster than the forward stroke. This is achieved by utilizing a specific gear ratio and mechanical design to reduce the time taken for the return phase.\n"
            "- **Adjustable Parameters:** Key parameters such as the radius of the pinions, the gear ratios, and the center distance between the racks can be adjusted to suit different applications and performance requirements.\n"
            "\n"
            "**Disadvantage**:\n"
            "One notable disadvantage of this mechanism is the potential for shock loading. This occurs due to the instantaneous change in direction at the ends of the strokes, where the mechanism transitions from forward to return motion and vice versa. This sudden change in speed can result in mechanical stress and vibrations, which may lead to wear and tear or reduced longevity of the components if not properly managed.\n"
            "\n"
            "**Applications**:\n"
            "This type of mechanism is intended to be used in the development levelling systems for rimless wheels.\n"
        )
        return description


import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from matplotlib.animation import FuncAnimation, PillowWriter


class mechanism14:
    def __init__(self, settings):
        self.settings = settings
        self.ani = None
        pass 

    @staticmethod
    def calculate_radius2(radius1, gear1_angle):
        radius2 = (gear1_angle * radius1) / (2 * np.pi - gear1_angle)
        return radius2
    
    @staticmethod
    def Settings(options, center_distance = 0, fig = None, ax = None, grid=False, rotations = 1, speed = 0.5):
        """
        Populate the settings dictionary with values from the options dictionary.

        Parameters:
        options (dict): A dictionary containing options and their values.

        Returns:
        dict: A dictionary containing the populated settings.
        """
        
        displacements = []
        radii_2 = []
        num_instances = len(options)
        origins = []
        phase_diff = 2*np.pi/num_instances
        for index, instance in enumerate(options):
            # if instance.get('offset_angle_degrees'):
            #     instance["offset_angle"] = np.radians(instance.get('offset_angle_degrees'))
            instance["offset_angle"] = np.radians(-90)
            offset_angle = instance.get('offset_angle')
            return_speed = instance.get('return_speed')
            return_angle = 2 * np.pi * 1/(return_speed + 1)
            forward_angle = 2 * np.pi - return_angle
            instance["forward_angle"] = forward_angle
            instance["phase"] = phase_diff * (index -1)
            instance["gear1_angle"] = forward_angle
            instance["start_angle_1"] = offset_angle + (2 * np.pi - forward_angle) / 2
            instance["stop_angle_1"] = instance.get('start_angle_1') + forward_angle
            instance["start_angle_2"] = instance["stop_angle_1"]
            instance["stop_angle_2"] =  instance["start_angle_2"] + return_angle

            # instance["radius2"] = mechanism14.calculate_radius2(instance.get('radius1', 0), forward_angle)
            instance["radius4"] =  forward_angle/ return_angle * instance["radius3"]
            radii_2.append(instance["radius1"]) ## fix
            displacements.append(forward_angle * instance.get('radius1', 0))
            instance["displacement"] = forward_angle * instance.get('radius1', 0)
            instance["x_limit"] = instance["displacement"]+10
            if center_distance == 0:
                center_distance = 2 * np.pi * instance.get('radius1', 0) * 2 
            origins.append(index * center_distance)

            

        settings = {"instances": options}
        if fig:
            settings['fig'] = fig 
        else:
            fig, ax = plt.subplots()
            ax.set_aspect('equal')
        settings['fig'] = fig 
        settings['ax'] = ax 

        # x_limit = sum(displacements) + (len(displacements) - 1) * center_distance
        x_limit = (len(options)) * center_distance + 10
        y_limit = 2 * max(radii_2) + 10

        settings["x_limit"] = x_limit
        settings["y_limit"] = y_limit


        ax.set_xlim(-x_limit/2, x_limit/2)
        ax.set_ylim(-y_limit/2, y_limit/2)
        ax.set_aspect('equal')

        if grid:
            plt.grid(True)

        if origins:
            mid_value = (max(origins) + min(origins)) / 2
            origins = [o - mid_value for o in origins]

        # Assign origins to each instance
        for index, instance in enumerate(options):
            instance["origin"] = origins[index]

        settings["rotations"] = rotations
        settings["speed"] = speed

        return settings

    def create_circle_section(self, r, num_points=100, start_angle=0, end_angle=2 * np.pi, phase=0):
        theta = np.linspace(start_angle+phase, end_angle+phase, num_points)
        x = r * np.cos(theta)
        y = r * np.sin(theta)
        return x, y

    def rack_displacement(self, angle, forward_angle, radius1, radius3, radius4):
        return_angle = 2 *np.pi  - forward_angle
        reversing_angle = angle
        if angle <= return_angle / 2:
            displacement = -angle * radius1 * (radius4/radius3)
            reversing_angle *= -1
        elif angle <= return_angle / 2 + forward_angle:
            last_position = -(return_angle / 2)* radius1 * (radius4/radius3)
            displacement = last_position + (angle - return_angle / 2) * radius1 
        else:
            last_position = (forward_angle / 2)* radius1
            displacement = last_position - (angle - (return_angle / 2 + forward_angle)) * radius1 * (radius4/radius3)
            reversing_angle *= -1
        
        return (displacement,reversing_angle)

    def plot(self, instance_number=None):
        fig = self.settings['fig']
        ax = self.settings['ax']

        if instance_number is not None:
            # Filter instances to only include the specified instance_number
            instances = [self.settings['instances'][instance_number]]
        else:
            # Plot all instances
            instances = self.settings['instances']

        for inst in instances:
            radius1 = inst.get('radius1', 0)
            forward_angle = inst.get('forward_angle', 0)
            start_angle_1 = inst.get('start_angle_1', 0)
            stop_angle_1 = inst.get('stop_angle_1', 0)
            start_angle_2 = inst.get('start_angle_2', 0)
            stop_angle_2 = inst.get('stop_angle_2', 0)
            origin = inst.get('origin', 0)
            x_limit = inst.get("x_limit")
            phase = inst.get("phase")

            # Plot the larger circle section
            x1, y1 = self.create_circle_section(radius1, start_angle=start_angle_1, end_angle=stop_angle_1, phase=phase)
            ax.plot(x1 + origin, y1, label=f'Larger Circle Section (Origin {origin})', color='blue', linestyle='--')

            # The return section
            x2, y2 = self.create_circle_section(radius1, start_angle=start_angle_2, end_angle=stop_angle_2, phase=phase)
            ax.plot(x2 + origin, y2, label=f'Smaller Circle Section (Origin {origin})', color='red', linestyle='--')

            # small pinion
            radius3 = inst.get('radius3', 0)
            x3, y3 = self.create_circle_section(radius3, start_angle=0, end_angle=2*np.pi, phase=phase)
            ax.plot(x3 + origin, y3-radius1-radius3, color='red', linestyle='--')

            # reversing pinion
            radius4 = inst.get('radius4', 0)
            x4, y4 = self.create_circle_section(radius4, start_angle=0, end_angle=2*np.pi, phase=phase)
            ax.plot(x4 + origin, y4-radius1-radius3, color='red', linestyle='--')

            angle = phase % (2*np.pi)
            x_end_circle1 = radius1 * np.cos(stop_angle_1 + angle)+origin
            y_end_circle1 = radius1 * np.sin(stop_angle_1 + angle)
            x_start_circle1 = radius1 * np.cos(start_angle_1 + angle)+origin
            y_start_circle1 = radius1 * np.sin(start_angle_1 + angle)

            ax.plot([x_start_circle1, x_end_circle1], [y_start_circle1, y_end_circle1], color='black', linestyle='-')

            # the rack
            rack_displacement, _= self.rack_displacement(angle, forward_angle, radius1, radius3, radius4)

            # # Add dashed horizontal lines
            ax.plot([-x_limit/2 + origin+rack_displacement, x_limit/2 + origin+rack_displacement], [-radius1, -radius1], color='blue', linestyle='--', label='Line at radius1')
            ax.plot([-x_limit/2 + origin+rack_displacement, x_limit/2 + origin+rack_displacement], [-radius1-radius3-radius4, -radius1-radius3-radius4], color='red', linestyle='--', label='Line at -radius2')

            # # Connect the starts and ends of the lines
            ax.plot([-x_limit/2 + origin+rack_displacement, -x_limit/2 + origin+rack_displacement], [-radius1, -radius1-radius3-radius4], color='black', linestyle='-', label='Connection Start')
            ax.plot([x_limit/2+origin+rack_displacement, x_limit/2+origin+rack_displacement], [-radius1, -radius1-radius3-radius4], color='black', linestyle='-', label='Connection End')


            # Add center dot
            ax.plot(origin, 0, 'ko')  # 'ko' stands for black color and circle marker
            ax.plot(origin,  -radius1-radius3, 'ko')  # 'ko' stands for black color and circle marker

        # Set plot settings
        ax.set_aspect('equal', adjustable='box')
        ax.set_title(f'mechanism 14 (n={len(instances)})')
        ax.set_xlabel('X')
        ax.set_ylabel('Y')
        ax.grid(True)
        ax.axis('off')
        
        

        return self


    def show(self):
        plt.show()
        return self
    
    def save(self, fileName):
        if self.ani is not None:
            writer = PillowWriter(fps=25)
            self.ani.save(f"{fileName}.gif", writer=writer)
        else:
            plt.savefig(f"{fileName}.png")
        return self
    
    def update(self, interval=100, frames=100):
        fig = self.settings['fig']
        ax = self.settings['ax']

        rotation = self.settings.get('rotation', 2 * np.pi)  # Default to 360 degrees (full rotation)
        speed = self.settings.get('speed', 1)  # Default speed factor

        # Calculate the total number of frames needed for the given rotation
        total_frames = int((rotation / (2 * np.pi)) * frames / speed)
        
        def animate(frame):
            ax.clear()  # Clear the current plot
            for inst in self.settings['instances']:
                radius1 = inst.get('radius1', 0)
                forward_angle = inst.get('forward_angle', 0)
                start_angle_1 = inst.get('start_angle_1', 0)
                stop_angle_1 = inst.get('stop_angle_1', 0)
                start_angle_2 = inst.get('start_angle_2', 0)
                stop_angle_2 = inst.get('stop_angle_2', 0)
                origin = inst.get('origin', 0)
                x_limit = inst.get("x_limit")
                radius3 = inst.get('radius3', 0)
                radius4 = inst.get('radius4', 0)
                phase = inst.get("phase")+ (frame * rotation / total_frames) 
                angle = phase % (2*np.pi)
                rack_displacement,reverse_angle = self.rack_displacement(angle, forward_angle, radius1, radius3, radius4)

                rack_displacement *= -1
                # Plot the larger circle section
                x1, y1 = self.create_circle_section(radius1, start_angle=start_angle_1, end_angle=stop_angle_1, phase=phase)
                x1 += rack_displacement
                ax.plot(x1 + origin, y1, label=f'Larger Circle Section (Origin {origin})', color='blue', linestyle='--')

                # The return section
                x2, y2 = self.create_circle_section(radius1, start_angle=start_angle_2, end_angle=stop_angle_2, phase=phase)
                x2 += rack_displacement
                ax.plot(x2 + origin, y2, label=f'Smaller Circle Section (Origin {origin})', color='red', linestyle='--')

                
                

                
                # small pinion
                x3, y3 = self.create_circle_section(radius3, start_angle=0, end_angle=2*np.pi, phase=reverse_angle)
                x3 += rack_displacement
                ax.plot(x3 + origin, y3-radius1-radius3, color='red', linestyle='--')
                # reversing pinion
                
                x4, y4 = self.create_circle_section(radius4, start_angle=0, end_angle=2*np.pi, phase=reverse_angle)
                x4 += rack_displacement
                ax.plot(x4 + origin, y4-radius1-radius3, color='red', linestyle='--')


                
                x_end_circle1 = radius1 * np.cos(stop_angle_1 + angle)+origin + rack_displacement
                y_end_circle1 = radius1 * np.sin(stop_angle_1 + angle)
                x_start_circle1 = radius1 * np.cos(start_angle_1 + angle)+origin + rack_displacement
                y_start_circle1 = radius1 * np.sin(start_angle_1 + angle)

                ax.plot([x_start_circle1, x_end_circle1], [y_start_circle1, y_end_circle1], color='black', linestyle='-')

                # the rack
                

                # # Add dashed horizontal lines
                ax.plot([-x_limit/2 + origin, x_limit/2 + origin], [-radius1, -radius1], color='blue', linestyle='--', label='Line at radius1')
                ax.plot([-x_limit/2 + origin, x_limit/2 + origin], [-radius1-radius3-radius4, -radius1-radius3-radius4], color='red', linestyle='--', label='Line at -radius2')

                # # Connect the starts and ends of the lines
                ax.plot([-x_limit/2 + origin, -x_limit/2 + origin], [-radius1, -radius1-radius3-radius4], color='black', linestyle='-', label='Connection Start')
                ax.plot([x_limit/2+origin, x_limit/2+origin], [-radius1, -radius1-radius3-radius4], color='black', linestyle='-', label='Connection End')


                # Add center dots
                ax.plot(origin+rack_displacement, 0, 'ko')
                ax.plot(origin+rack_displacement,  -radius1-radius3, 'ko')  

            ax.set_aspect('equal', adjustable='box')
            ax.set_title(f'mechanism 14 (n={len(self.settings["instances"])})')
            ax.set_xlabel('X')
            ax.set_ylabel('Y')
            ax.grid(True)
            x_limit = self.settings["x_limit"]
            y_limit = self.settings["y_limit"]
            self.settings["x_limit"] = x_limit
            self.settings["y_limit"] = y_limit


            ax.set_xlim(-x_limit/2, x_limit/2)
            ax.set_ylim(-y_limit/2, y_limit/2)
            ax.axis('off')

        self.ani = animation.FuncAnimation(fig, animate, frames=total_frames, interval=interval, repeat=True)
        return self
    
    @staticmethod
    def info():
        description = (
            "Mechanism Description:\n"
            "-----------------------\n"
            "This mechanism is a reciprocating linear motion system featuring constant velocity and a quick return mechanism. "
            "It is constructed using two racks and a compound mangled pinion and a compound gear, designed to convert continuous rotary motion into linear motion with specific characteristics:\n"
            "\n"
            "- **Constant Velocity:** The mechanism ensures a uniform linear speed during the forward stroke, providing smooth and consistent motion.\n"
            "- **Quick Return:** It incorporates a quick return feature, which means the return stroke (or reverse movement) is faster than the forward stroke. This is achieved by utilizing a specific gear ratio and mechanical design to reduce the time taken for the return phase.\n"
            "- **Adjustable Parameters:** Key parameters such as the radius of the pinions, the gear ratios, and the center distance between the racks can be adjusted to suit different applications and performance requirements.\n"
            "\n"
            "**Disadvantage**:\n"
            "One notable disadvantage of this mechanism is the potential for shock loading. This occurs due to the instantaneous change in speed at the ends of the strokes, where the mechanism transitions from forward to return motion and vice versa. This sudden change in speed can result in mechanical stress and vibrations, which may lead to wear and tear or reduced longevity of the components if not properly managed.\n"
            "\n"
            "**Applications**:\n"
            "This type of mechanism is intended to be used in the development of clamp-on self-actuated long-range linear motion mechanisms.\n"
        )
        return description


import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation

from ....utils import bmutils
import time
class mechanism32(bmutils):
    def __init__(self, settings):
        self.settings = settings
        self.fig, self.ax = plt.subplots(figsize=(10, 6))
        self.ani = None

        # Extract parameters from settings
        self.R1 = self.settings['R1']
        self.R2 = self.settings['R2']
        self.R3 = self.settings['R3']
        self.R4 = self.settings['R4']
        self.R_S = self.settings['R_S']
        self.R_A = self.settings['R_A']
        self.R_B = self.settings['R_B']
        self.R_External = self.settings['R_External']
        self.theta_D = self.settings['theta_D']
        # self.theta_forward = self.settings['theta_forward']
        # self.previous_theta = 0
        self.instances = settings["instances"]
        self.phase_6 = settings["phase_6"]
        self.loop_index= 0
        self.footpath = {}
        
        self.setup_plot()

    def setup_plot(self):
        self.ax.set_title(f'Mechanism 32')
        linear_rack_length = self.settings["linear_rack_length"]
        self.ax.set_xlim(-linear_rack_length/2, linear_rack_length/2)
        self.ax.set_ylim(-self.R_B*2-self.settings["leg_length"], self.R_B*2)
        self.ax.axhline(0, color='black', linewidth=0.5, linestyle='--')
        # self.ax.grid()
        self.ax.axis("off")
        self.ax.set_aspect('equal')
        super().watermark(self.ax)
    
    def plot(self, theta = 0):    
        self.setup_plot()
        # the racks (fixed internal gears)
        x_A, y_A = self.create_circle_section(self.R_A)
        self.ax.plot(x_A, y_A, color='black', linestyle='-.')
        self.ax.plot([0], [0], 'ro') 

        if theta == 0:
            self.loop_index = 0


        x1, y1 = self.create_circle_section(self.R1, phase=theta)
        self.ax.plot(x1, y1, color='red', linestyle='dotted')
        x1, y1 = self.create_circle_section(self.R1, phase=theta, end_angle=0.2*np.pi) # marker (solid section)
        self.ax.plot(x1, y1, color='red', linestyle='-', linewidth=5)
      
        # check descriptions of these in settings method
        R1 = self.R1
        R2 = self.R2
        R3 = self.R3
        R4 = self.R4
        R_S = self.R_S
        R_A = self.R_A
        R_B = self.R_B
        R_External = self.R_External
       
        segment_instances = self.settings["segment_instances"]

        def get_previous_section_type(prev_theta, theta_2_initial = 0):
            sections_max_thetas =[
                [segment_instance["section_max_theta"] for segment_instance in segment_instances if segment_instance["type"] == 0][0], # since we are starting from theta = 0
                [segment_instance["section_max_theta"] for segment_instance in segment_instances if segment_instance["type"] != 0][0],
            ]
            section_types = [
                [segment_instance["type"] for segment_instance in segment_instances if segment_instance["type"] == 0][0],
                [segment_instance["type"] for segment_instance in segment_instances if segment_instance["type"] != 0][0],
            ]
            rack_theta = 0
            d_theta_2 = R1/R2 * prev_theta
            current_section_type = 0
            # previous_section_type = 0
            previous_section_theta = 0
            index_while = 0
            while d_theta_2 > 0 : # we are currently only working with positive d_thetas
                binary_index = (index_while)%2
                section_theta_max = sections_max_thetas[binary_index]
                remaining_for_previous_section =  section_theta_max - previous_section_theta
                used_in_section = d_theta_2 if d_theta_2 <= remaining_for_previous_section else remaining_for_previous_section
                
                total_section_theta = previous_section_theta + used_in_section
                previous_section_theta = total_section_theta if total_section_theta < section_theta_max else 0

                
                binary_index_next = (index_while+1)%2
                current_section_type = section_types[binary_index]

                if current_section_type == 0: 
                    active_rack_r = R_B
                    active_planet_r = R2
                else:
                    active_rack_r = R_S
                    active_planet_r = R4
                    
                d_theta_rack = used_in_section * R3/R_A - used_in_section * active_planet_r/active_rack_r #
                rack_theta += d_theta_rack   
                
                d_theta_2 -= used_in_section
                index_while += 1
            theta_rack_no_offset = rack_theta
            rack_theta -= theta_2_initial * R3/R_A 
            return rack_theta, theta_rack_no_offset
    
        def theta_2_to_theta_1(theta_2): # will be same value if R2 = R1
            return theta_2 * R2/R1
        
        midpoints = []
        colors = []

        for index, instance in enumerate(self.instances):
            self.__dict__.update(instance)
            
            self.theta = theta_2_to_theta_1(self.theta_2_init) + theta
            self.theta_rack, theta_rack_no_offset = get_previous_section_type(self.theta, theta_2_to_theta_1(self.theta_2_init)) # what bug is it here??? use so that we do not need to set theta_rack_init and previous_section_type_init
            
            '''
            Driver parts
            '''
            R2_center =   [R1+R2  , 0]
            
            centerx, centery = self.rotate_points(self.phase_6, [R2_center[0]],[R2_center[1]], [0,0])
            R2_center = [centerx[0],centery[0]]
            # centerx, centery = self.rotate_points(self.phase_6, [R_intersect[0]],[R_intersect[1]], [0,0])
            # R_intersect = [centerx[0],centery[0]]

            theta_2 = - R1/R2 * self.theta # self.theta is theta of R1
            theta_A = -theta_2 * R3 / R_A #+ self.theta_D/2 # total theta_A (Annulus)
            centerx, centery = self.rotate_points(theta_A, [R2_center[0]],[R2_center[1]], [0,0]) # it is starting at the wrong
            R2_center = [centerx[0],centery[0]]
            if index == 0:
                self.ax.plot([R2_center[0]], [R2_center[1]], 'ro')

            # centerx, centery = self.rotate_points(theta_A, [R_intersect[0]],[R_intersect[1]], [0,0]) # on fixed annulus. 
            # R_intersect = [centerx[0],centery[0]]
            # self.ax.plot([R_intersect[0]], [R_intersect[1]], 'ro')
           
            if index == 0:
                x2, y2 = self.create_circle_section(self.R2, phase=theta_2)
                x3, y3 = self.create_circle_section(self.R3, phase=theta_2)
                self.ax.plot(x2+R2_center[0], y2+R2_center[1], color='red', linestyle='dotted')
                self.ax.plot(x3+R2_center[0], y3+R2_center[1], color='black', linestyle='dotted')


            # if instance["type"] == 0:
            if index == 0:
                self.ax.plot(x2+R2_center[0], y2+R2_center[1], color='green', linestyle=':')
                x2, y2 = self.create_circle_section(self.R2, phase=theta_2, end_angle=0.1*np.pi)
                self.ax.plot(x2+R2_center[0], y2+R2_center[1], color='green', linestyle='-', linewidth=4)
                # else:           
                x4, y4 = self.create_circle_section(self.R4, phase=theta_2)
                self.ax.plot(x4+R2_center[0], y4+R2_center[1], color='blue', linestyle='--')
                x2, y2 = self.create_circle_section(self.R2, phase=theta_2, end_angle=0.1*np.pi)
                self.ax.plot(x2+R2_center[0], y2+R2_center[1], color='red', linestyle='-', linewidth=4)
            '''
            End of driver parts
            '''

            for key, value in self.__dict__.items():
                if key in instance:  # Check if the key exists in the instance
                    instance[key] = value  # Update the dictionary
                elif hasattr(instance, '__dict__') and key in vars(instance):
                    setattr(instance, key, value)

            
            '''
            Driven parts (output)
            '''

            # ''' Markers '''

            segment_endpoints = []
            for _instance in segment_instances:
                color = None
                if "color" in instance.keys():
                    color = instance["color"]
            
                if _instance["type"] == 0:
                    active_r = self.R_B
                    color = color if color else ("green" if index == 0 else "orange")
                else:
                    active_r = self.R_S
                    color = color if color else ("green" if index == 0 else "orange")

                x_i, y_i = self.create_circle_section(active_r, phase=self.phase_6+self.theta_rack + _instance["theta"], end_angle=self.theta_D)
                instance["segment"] = [x_i, y_i]
                self.ax.plot(x_i, y_i, color=color, linestyle='--',)
                
                segment_endpoints.append([x_i[-1], y_i[-1]])
                segment_endpoints.append([x_i[0], y_i[0]])

            for index_, point in enumerate(segment_endpoints): # the lines conneting end of R_B to beginning of R_S
                if index_%2 == 0:
                    continue
                next_point_index = index_+1 if (index_+1) < len(segment_endpoints) else 0
                next_point = segment_endpoints[next_point_index]

                # Draw lines between the points
                self.ax.plot([point[0],next_point[0]],[point[1],next_point[1]] , color=color) 


            hip_distance = self.settings["hip_distance"]
            hip_distance = -hip_distance/2 if index%2 == 0 else + hip_distance/2
            self.linear_rack_displacement = (-R_External * theta_rack_no_offset) + hip_distance
            # self.ax.plot([self.linear_rack_displacement], [0], color=color)
            linear_rack_len = self.settings["linear_rack_length"]
            max_rack_displacement = self.settings["max_rack_displacement"]
            # x_left = -linear_rack_len/2+self.linear_rack_displacement - self.linear_rack_displacement/2
            # x_right = linear_rack_len/2+self.linear_rack_displacement - self.linear_rack_displacement/2
            midpoint = -self.linear_rack_displacement + linear_rack_len/2
            midpoint = +self.linear_rack_displacement - linear_rack_len/2

            
            midpoint = -self.linear_rack_displacement + max_rack_displacement/2 
            
            x_left = midpoint - linear_rack_len/2
            x_right = midpoint + linear_rack_len/2

            midpoints.append(midpoint)
            colors.append(color)
            
            self.ax.plot([midpoint], [-R_External], 'o',  color=color) 
            self.ax.plot([[x_left],[x_right]],[[-R_External],[-R_External]], color=color) # racks

            x_External, y_External = self.create_circle_section(R_External, phase=self.phase_6+self.theta_rack, end_angle=0.1*np.pi)
            self.ax.plot(x_External, y_External, color=color, linestyle='-', linewidth=5)
            # break
            
        x_External, y_External = self.create_circle_section(R_External, phase=self.phase_6+self.theta_rack)
        self.ax.plot(x_External, y_External, color='black',linestyle='--')


        def find_intersection(point0, point1, r1, R_External):
            # Calculate the midpoint for x
            x_intersect = (point0[0] + point1[0]) / 2
            
            # Solve for y
            y_intersect = -R_External - np.sqrt(r1**2 - (x_intersect - point0[0])**2)
            
            return (x_intersect, y_intersect)

        r1 = self.settings["leg_length"]
        for i in range(0, len(midpoints), 2):
            group = midpoints[i:i+2]
            
            point0 = group[0]
            point1 = group[1]
            point0_ = point0
            point1_ = point1
            theta1 = np.linspace(0, np.pi, 100)
            x1 = point0+ r1 * np.cos(theta1)
            y1 = -R_External-r1 * np.sin(theta1)

            # Create semicircle 2
            theta2 = np.linspace(0, np.pi, 100)
            x2 = point1+ + r1 * np.cos(theta2)
            y2 = -R_External-r1 * np.sin(theta2)

            # get point of intersection between the 2 semicircles.

            point0 = np.array([point0, -R_External])  # Coordinates of point0
            point1 = np.array([point1, -R_External])  # Coordinates of point1

            intersection = find_intersection(point0, point1, r1, R_External)

            # Plotting the intersection point
            color = colors[i]
            xs = [[point0_],[intersection[0]]]
            ys = [ [-R_External],[intersection[1]]]
            # self.ax.plot([],[])  
            self.ax.plot(xs, ys, color=color)
            color = colors[i+1]
            xs = [[point1_],[intersection[0]]]  
            self.ax.plot(xs, ys, color=color)  

            # self.ax.plot([[x_left],[x_right]],[[-R_External],[-R_External]], color=color) # racks
            self.ax.plot(intersection[0], intersection[1], 'ro', label='Intersection Point')  
            if i == 0:
                j = self.loop_index
                if j not in self.footpath:  # Only add if there's no intersection at this index
                    self.footpath[j] = intersection
        
        keys = sorted(self.footpath.keys())  # Get sorted keys to maintain order
        for j in range(len(keys) - 1):
            x_values = [self.footpath[keys[j]][0], self.footpath[keys[j + 1]][0]]
            y_values = [self.footpath[keys[j]][1], self.footpath[keys[j + 1]][1]]
            self.ax.plot(x_values, y_values, 'b--')  # Draw a blue line between intersections
        self.loop_index += 1
        return self
    
    def update(self, interval=100, frames=100):
        speed = self.settings["speed"]
        len = self.settings["animation_rounds"]
        displacements = np.linspace(0, len, int(250/speed))
        # displacements = np.concatenate([displacements, displacements[::-1]])  # Reverse motion
        def animate(displacement):
            self.ax.clear()
            self.plot(displacement)

        
        self.setup_plot()
        self.ani = animation.FuncAnimation(self.fig, lambda displacement: animate(displacement), frames=displacements, interval=50, repeat=True)
        return self
        

    @staticmethod
    # return_speed: how many times return is faster than forward
    # frequency: number of oscillations of output per revolution of planet. Should be greater than one
    def Settings(R1=5,R2=20, return_speed=2, frequency = 1, num_legs =1, phase_difference = 0, animation_rounds=1, speed=1, phase_difference_for_legs=np.pi,hip_distance=None):
        num_instances = num_legs * 2
        phase_difference_fraction = phase_difference/2*np.pi
        if not hip_distance:
            hip_distance = R2
        num_hands = frequency *2
        theta_segment = 2*np.pi/num_hands # angle of rack segment used
        theta_D = theta_segment # theta_disk

        k = return_speed

        R_B = R1 + 2 * R2 # R_Big
        theta_2_R = theta_D * R_B / R2  # Disk 2 rotates through theta_2_R to cover distance of theta_D * R_B(entire length of rack)

        theta_D_R  = (1/(k+1) ) * theta_D
        theta_D_F = 2 * theta_D - theta_D_R

        theta_3_R = theta_2_R
        c = theta_D_R / theta_3_R
        R3 = (c * R1 + c * R2)/(1-c)
        R_A = R1 + R2 + R3  # R_Annulus
        
        theta_3_F = theta_D_F * R_A / R3
        theta_4_F = theta_3_F
        c = theta_D / theta_4_F

        R4 = c*(R1 + R2) / (1-c)
        R_S = R1 + R2 + R4 # R_Small. Moves mechanism forward in contrast to R_Big which moves it backwards

        R_External = R_B * 1.2

        R_CAM = R_B * 1.4

        theta_3_perimeter = 2*np.pi * R_A / R3
        s_2_perimeter = theta_3_perimeter * R2
        theta_1_perimeter = s_2_perimeter / R1
        perimeter_theta = theta_1_perimeter
        animation_rounds = int(animation_rounds * perimeter_theta)
        theta_points = [(perimeter_theta/num_hands) * i for i in range(num_hands)]
        
        # for sections on R_B(ig) and R_S(mall)
        s_section_B = theta_D * R_B # this is theta 2 since s_section_B x R2 = theta_D * R_B
        s_section_S = theta_D * R_S
        theta_max_B = s_section_B/R2 # both theta2 and theta3 since they are equal, being co-axial
        theta_max_S = s_section_S/R4 # both theta4 and theta2
        section_max_theta = [theta_max_B,theta_max_S ] # theta_3s
        theta_CAM = 0.75 * theta_D_F


        # how to calculate theta_2_init from phase_difference_fraction 
        R_A_difference = phase_difference_fraction
        fraction = 0.5
        # print(theta_max_B*R2/R_A)
        theta_3_init = R_A * phase_difference_fraction/R3
        theta_2_init = theta_3_init

        phase_difference_for_legs = R_A * phase_difference_for_legs/R3

        linear_rack_length = 1.5 * R_External * theta_D
        max_rack_displacement =  R_External * theta_D
        # instances = [
        #     {"theta":0, "theta_rack":0, "theta_2_init":0, "linear_rack_displacement":0,"color":"black"},
        #     {"theta":0, "theta_rack":0, "theta_2_init":theta_2_init, "linear_rack_displacement":0, "color":"orange"},
        #     # {"theta":0, "theta_rack":0, "theta_2_init":0, "color":"black"},
        #     # #  {"theta":0, "theta_rack":0, "theta_init":0, "color":"black"},
            
        #     # {"theta":s_section_S*fraction, "theta_init":s_section_S*fraction, "theta_rack":-theta_D*fraction,"theta_rack_init":-theta_D*fraction}
        # ] # the second one starts at end of first B
        colors = [
           "black","orange", "green", "gray"
        ]
        instances = []
        # if num_instances > 2:
        # for i in range(2,num_instances):
        for i in range(num_instances):
            instance = {} # instances[i%2].copy()
            phase = (i // 2) 
            initial_angle = theta_2_init * (i%2) + phase_difference_for_legs * phase
            instance["theta_2_init"]= initial_angle
            # instance["theta_2_init"] = theta_2_init * (i)
            instance["color"] = colors[i%len(colors)]
            instance["theta_rack"] = 0
            instance["theta"] = 0
            instance["linear_rack_displacement"] = 0
            instances.append(instance)

        segment_instances = [{"theta":theta, "type":index%2, "section_max_theta":section_max_theta[index%2]} for index, theta in enumerate(theta_points)]
        # for index, instance in enumerate(instances):
        #     instance["previous_direction"] = 1

        phase_6 = - theta_D_R/2 - np.pi/2
        return {
            "R1": R1,
            "R2": R2,
            "R3": R3,
            "R4": R4,
            "R_S": R_S,
            "R_A": R_A,
            "R_B": R_B,
            "R_External": R_External,
            "R_CAM": R_CAM,
            "theta_CAM": theta_CAM,
            "theta_D": theta_D,
            "leg_length": 2*R_External,
            # "theta_forward":theta_forward,
            "animation_rounds":animation_rounds,
            "speed":speed,
            "perimeter_theta":perimeter_theta,
            "linear_rack_length":linear_rack_length,
            "max_rack_displacement":max_rack_displacement,
            "hip_distance":hip_distance, # distance between 2 hips of single leg
            # "theta_limits":theta_limits,
            "instances":instances,
            "segment_instances":segment_instances,
            "phase_6":phase_6,
            "frequency":frequency,
        }
    @staticmethod
    def info():
        description = (
            "Mechanism Description:\n"
            "-----------------------\n"
            "Walking Mechanism with uz = 0 and ux = k.\n"
            "**Applications**:\n"
            "This type of mechanism is intended to be used in the development of clamp-on self-actuated long-range linear motion mechanisms and walking mechanisms.\n"
        )
        return description


import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation

class mechanism22:
    def __init__(self, settings):
        self.settings = settings
        self.fig, self.ax = plt.subplots(figsize=(10, 6))
        self.ani = None

        # Extract parameters from settings
        self.R1 = self.settings['R1']
        self.R2 = self.settings['R2']
        self.R3 = self.settings['R3']
        self.L = self.settings['L']
        self.theta_R = self.settings['theta_R']
        self.theta_forward = self.settings['theta_forward']
        self.centers = self.settings['centers']
        self.displacement = self.settings['displacement']

        self.mobile_rack_xes = [-self.L,0]
        self.active_wheel_index = 0
        self.previous_active_wheel_index = 0
        self.previous_theta = 0
        self.cycle_theta = 0
        self.stroke_start_theta = 0

        self.setup_plot()

    def setup_plot(self):
        # max_radius = max(inst['radius1'] for inst in self.instances)
        self.ax.set_xlim(-self.L * 4, 4 * self.L)
        self.ax.set_ylim(-2 * self.R3, 2 * self.R3)
        self.ax.axhline(0, color='black', linewidth=0.5, linestyle='--')
        self.ax.grid()
        self.ax.set_aspect('equal')

    def create_circle_section(self, r, num_points=100, start_angle=0, end_angle=2 * np.pi, phase=0):
        theta = np.linspace(start_angle + phase, end_angle + phase, num_points)
        x = r * np.cos(theta)
        y = r * np.sin(theta)
        return x, y

    def plot(self, displacement=0):        
        self.ax.set_title(f'Mechanism 22')

        centers = [i * self.L - displacement for i in range(len(self.centers))]
        theta_R2 = displacement/self.R2
        x2, y2 = self.create_circle_section(self.R2, phase=theta_R2)
        x3, y3 = self.create_circle_section(self.R3, phase=theta_R2)
        x1, y1 = self.create_circle_section(self.R1, phase=theta_R2)
        
        for i,center in enumerate(centers):
            if i%2 ==0:
                self.ax.plot(x2+center, y2, color='black', linestyle='--')
                self.ax.plot(x3+center, y3, color='blue', linestyle='--')
            else:
                self.ax.plot(x2+center, y2, color='black', linestyle='--')
                self.ax.plot(x1+center, y1, color='green', linestyle='--')

            if self.mobile_rack_xes[0] <= center and center <= self.mobile_rack_xes[1]:
                self.active_wheel_index = i

        if self.previous_active_wheel_index != self.active_wheel_index:
            self.stroke_start_theta = theta_R2
            self.previous_active_wheel_index = self.active_wheel_index
            self.previous_theta = theta_R2
        # cycle_theta = theta_R2 - self.stroke_start_theta

        d_theta = theta_R2 - self.previous_theta
        self.previous_theta = theta_R2
        
        stroke_displacement = 0
        if self.active_wheel_index % 2 == 0:
            stroke_displacement = -((d_theta * self.R2) - d_theta * self.R3)
        else:
            stroke_displacement = -((d_theta * self.R2) - d_theta * self.R1)
        self.mobile_rack_xes[0] += stroke_displacement
        self.mobile_rack_xes[1] += stroke_displacement

        self.ax.plot([-self.L*4, self.L*4], [-self.R2, -self.R2], color='black', linestyle='--')
        
        # Set limits
        self.ax.set_ylim(-self.R3*2, self.R3*2)  # Adjust based on your specific needs
        self.ax.set_xlim(-self.L*4, 4*self.L)  # Adjust limits based on L

        # Annotations
        self.ax.set_title('Mechanism 22')
        self.ax.axis('off')
        self.ax.grid()
        self.ax.axhline(0, color='black', linewidth=0.5, linestyle='--')
        self.ax.plot(centers, [0] * len(centers), 'ro', label='Centers')  # Red dots at the centers
        self.ax.plot(centers, [0] * len(centers), 'r--')  # Red line through the centers

        mobile_rack_xes = self.mobile_rack_xes
        self.ax.plot(mobile_rack_xes, [-self.R3, -self.R3], color='blue', linestyle='--')
        self.ax.plot(mobile_rack_xes, [-self.R1, -self.R1], color='green', linestyle='--')
        self.ax.plot([mobile_rack_xes[0], mobile_rack_xes[0]], [-self.R1, -self.R3], color='pink', linestyle='--')
        self.ax.plot([mobile_rack_xes[1], mobile_rack_xes[1]], [-self.R1, -self.R3], color='pink', linestyle='-')
        return self

    def update(self, interval=100, frames=100):
        max = (self.settings["num_wheels"]-1)*150 /6
        max = int(max)
        displacements = np.linspace(0, (self.settings["num_wheels"]-1)*self.L, max)
        displacements = np.concatenate([displacements, displacements[::-1]])  # Reverse motion
        def animate(displacement):
            self.ax.clear()
            self.plot(displacement)

        
        self.setup_plot()
        self.ani = animation.FuncAnimation(self.fig, lambda displacement: animate(displacement), frames=displacements, interval=50, repeat=True)
        return self

    def show(self):
        plt.show()
        return self

    def save(self, fileName):
        if self.ani is not None:
            # writer = PillowWriter(fps=25)
            self.ani.save(f"{fileName}.gif", writer='Pillow') # how to set fps
        else:
            plt.savefig(f"{fileName}.png")
        return self

    @staticmethod
    def Settings(R2=3, R3=4, L=10, num_wheels=4, displacement=0):
        
        theta_R = L / R3
        # theta_forward = 2 * theta_R * (R3 / R2 - 1)
        theta_forward = theta_R * (2*R3 / R2 - 1)
        # Avoid division by zero for R1
        if theta_forward != 0:
            R1 = L / theta_forward
        else:
            R1 = np.nan
        
        # Horizontal dashed line from -L/2 to +L/2
        centers = [i * L for i in range(num_wheels)]
        return {
            "R1": R1,
            "R2": R2,
            "R3": R3,
            "L": L,
            "theta_R": theta_R,
            "theta_forward": theta_forward,
            "centers": centers,
            "displacement": displacement,
            "num_wheels": num_wheels,
        }
    
    @staticmethod
    def info():
        description = (
            "Mechanism Description:\n"
            "-----------------------\n"
            "This mechanism is a continuous linear motion into reciprocating linear motion.\n"
            "**Applications**:\n"
            "This type of mechanism is intended to be used in the development of clamp-on self-actuated long-range linear motion mechanisms.\n"
        )
        return description
